import pytest
import numpy as np
from arvi.binning import bin_ccf_mask


def test_bin_ccf_mask_basic():
    time = np.array([1.1, 1.2, 2.1, 2.2, 2.3])
    ccf_mask = np.array(['A', 'A', 'B', 'B', 'B'])
    expected = np.array(['A', 'B'])
    result = bin_ccf_mask(time, ccf_mask)
    assert np.array_equal(result, expected)


def test_bin_ccf_mask_nonunique():
    time = np.array([1.1, 1.2, 1.3, 2.1, 2.2])
    ccf_mask = np.array(['A', 'B', 'A', 'B', 'B'])
    expected = np.array(['nan', 'B'])
    result = bin_ccf_mask(time, ccf_mask)
    assert np.array_equal(result, expected)


def test_bin_ccf_mask_empty():
    time = np.array([])
    ccf_mask = np.array([])
    result = bin_ccf_mask(time, ccf_mask)
    assert result.size == 0



import pytest
import numpy as np
from arvi.stats import wmean


def test_wmean():
    a = np.array([1, 2, 3])
    e = np.array([0.1, 0.2, 0.3])
    expected = (a/e**2).sum() / (1/e**2).sum()
    result = wmean(a, e)
    assert result == pytest.approx(expected)


def test_wmean_zero_weights():
    a = np.array([1, 2, 3])
    e = np.array([0, 0, 0])
    with pytest.raises(ZeroDivisionError):
        wmean(a, e)


def test_wmean_invalid_inputs():
    # Incompatible shapes
    a = np.array([1, 2, 3])
    e = np.array([0.1, 0.2])
    with pytest.raises(ValueError):
        wmean(a, e)

    # Negative uncertainties
    a = np.array([1, 2, 3])
    e = np.array([-0.1, 0.2, 0.3])
    with pytest.raises(ValueError):
        wmean(a, e)



from functools import partialmethod, wraps
from itertools import cycle

import numpy as np

from astropy.timeseries import LombScargle

from .setup_logger import setup_logger
from .config import config
from .stats import wmean

from .utils import lazy_import
plt = lazy_import('matplotlib.pyplot')


def plot_settings(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        # with plt.style.context('fast'):
        theme = 'dark_background' if config.dark_plots else 'fast'
        with plt.style.context(theme):
            return func(*args, **kwargs)
    return wrapper



class BlittedCursor:
    """ A cross-hair cursor using blitting for faster redraw. """
    def __init__(self, axes, vertical=True, horizontal=True, show_text=None,
                 transforms_x=None, transforms_y=None):
        import matplotlib # delay import
        if isinstance(axes, matplotlib.axes.Axes):
            axes = [axes]
        self.axes = axes
        self.background = None
        self.vertical = vertical
        self.horizontal = horizontal
        
        self.transforms_x = [lambda x:x for _ in axes] if transforms_x is None else transforms_x
        self.transforms_y = [lambda x:x for _ in axes] if transforms_y is None else transforms_y

        if horizontal:
            self.horizontal_line = [ax.axhline(color='k', lw=0.8, ls='--') for ax in axes]
        if vertical:
            self.vertical_line = [ax.axvline(color='k', lw=0.8, ls='--') for ax in axes]

        self.show_text = show_text
        if show_text is not None: # text location in axes coordinates
            self.text = [ax.text(0.72, 0.9, '', transform=ax.transAxes) for ax in axes]

        self._creating_background = False
        for ax in axes:
            ax.figure.canvas.mpl_connect('draw_event', self.on_draw)

    def on_draw(self, event):
        self.create_new_background()

    def set_cross_hair_visible(self, visible):
        if self.horizontal:
            need_redraw = [line.get_visible() != visible for line in self.horizontal_line]
        else:
            need_redraw = [line.get_visible() != visible for line in self.vertical_line]
        if self.horizontal:
            [line.set_visible(visible) for line in self.horizontal_line]
        if self.vertical:
            [line.set_visible(visible) for line in self.vertical_line]
        if self.show_text:
            self.text.set_visible(visible)
        return need_redraw

    def create_new_background(self):
        if self._creating_background:
            # discard calls triggered from within this function
            return
        self._creating_background = True
        self.set_cross_hair_visible(False)
        for ax in self.axes:
            ax.figure.canvas.draw()
        self.backgrounds = [ax.figure.canvas.copy_from_bbox(ax.bbox) for ax in self.axes]
        self.set_cross_hair_visible(True)
        self._creating_background = False

    def on_mouse_move(self, event):
        if self.background is None:
            self.create_new_background()
        if not event.inaxes:
            need_redraw = self.set_cross_hair_visible(False)
            if any(need_redraw):
                for ax, bkgd in zip(self.axes, self.backgrounds):
                    ax.figure.canvas.restore_region(bkgd)
                    ax.figure.canvas.blit(ax.bbox)
        else:
            self.set_cross_hair_visible(True)
            # update the line positions
            x, y = event.xdata, event.ydata
            X = [trans(x) for trans in self.transforms_x]
            Y = [trans(y) for trans in self.transforms_y]
            if self.horizontal:
                [line.set_ydata([y]) for line, y in zip(self.horizontal_line, Y)]
            if self.vertical:
                [line.set_xdata([x]) for line, x in zip(self.vertical_line, X)]
            if self.show_text:
                self.text.set_text(f'x={x:1.2f}, y={y:1.2f}')

            for ax, bkgd in zip(self.axes, self.backgrounds):
                ax.figure.canvas.restore_region(bkgd)
                if self.horizontal:
                    [ax.draw_artist(line) for line in self.horizontal_line]
                if self.vertical:
                    [ax.draw_artist(line) for line in self.vertical_line]
                if self.show_text:
                    ax.draw_artist(self.text)
                ax.figure.canvas.blit(ax.bbox)

def clickable_legend(fig, ax, leg):
    from matplotlib.text import Text
    handles, labels = ax.get_legend_handles_labels()
    for text in leg.get_texts():
        text.set_picker(True)

    def on_pick_legend(event):
        artist = event.artist
        if isinstance(artist, Text):
            try:
                h = handles[labels.index(artist.get_text())]
                alpha_text = {None:0.2, 1.0: 0.2, 0.2:1.0}[artist.get_alpha()]
                alpha_point = {None: 0.0, 1.0: 0.0, 0.2: 1.0}[artist.get_alpha()]
                try:
                    h[0].set_alpha(alpha_point)
                    h[2][0].set_alpha(alpha_point)
                except TypeError:
                    h.set_alpha(alpha_point)

                artist.set_alpha(alpha_text)
                fig.canvas.draw()
            except ValueError:
                pass
    return on_pick_legend

@plot_settings
def plot(self, ax=None, show_masked=False, instrument=None, time_offset=0,
         remove_50000=False, tooltips=True, show_title=False, show_legend=True, label=None, 
         jitter=None, N_in_label=False, versus_n=False, show_histogram=False, bw=False, **kwargs):
    """ Plot the RVs

    Args:
        ax (Axes, optional):
            Axis to plot to. Defaults to None.
        show_masked (bool, optional):
            Show masked points. Defaults to False.
        instrument (str, optional):
            Which instrument to plot. Defaults to None, or plot all instruments.
        time_offset (int, optional):
            Value to subtract from time. Defaults to 0.
        remove_50000 (bool, optional):
            Whether to subtract 50000 from time. Defaults to False.
        tooltips (bool, optional):
            Show information upon clicking a point. Defaults to True.
        show_title (bool, optional):
            Show the star name in the plot title. Defaults to False.
        show_legend (bool, optional):
            Show legend. Defaults to True.
        N_in_label (bool, optional):
            Show number of observations in legend. Defaults to False.
        versus_n (bool, optional):
            Plot RVs vs observation index instead of time. Defaults to False.
        show_histogram (bool, optional)
            Whether to show a panel with the RV histograms (per intrument).
            Defaults to False.
        bw (bool, optional):
            Adapt plot to black and white. Defaults to False.

    Returns:
        Figure: the figure
        Axes: the axis
    """
    logger = setup_logger()
    if self.N == 0:
        if self.verbose:
            logger.error('no data to plot')
        return

    if ax is None:
        if show_histogram:
            fig, (ax, axh) = plt.subplots(1, 2, constrained_layout=True, gridspec_kw={'width_ratios': [3, 1]})
        else:
            fig, ax = plt.subplots(1, 1, constrained_layout=True)
    elif ax == -1:
        ax = plt.gca()
        fig = ax.figure
    else:
        if show_histogram:
            ax, axh = ax
        fig = ax.figure

    kwargs.setdefault('ls', '')
    kwargs.setdefault('capsize', 0)
    kwargs.setdefault('ms', 4)

    if remove_50000:
        time_offset = 50000

    strict = kwargs.pop('strict', False)
    instruments = self._check_instrument(instrument, strict=strict)
    marker = kwargs.pop('marker', 'o')

    if bw:
        markers = cycle((marker, 'P', 's', '^', '*'))
    else:
        markers = cycle((marker,) * len(instruments))

    try:
        zorders = cycle(-np.argsort([getattr(self, i).error for i in instruments])[::-1])
    except AttributeError:
        zorders = cycle([1] * len(instruments))

    containers = {}
    for _i, inst in enumerate(instruments):
        s = self if self._child else getattr(self, inst)
        if s.mask.sum() == 0:
            continue

        if label is None:
            _label = f'{inst:10s} ({s.N})' if N_in_label else inst
            if not self.only_latest_pipeline:
                i, p = _label.split('_', 1)
                p = p.replace('_', '.')
                _label = f'{i}-{p}'
        else:
            if isinstance(label, list):
                _label = label[_i]
            else:
                _label = label

        if versus_n:
            x = np.arange(1, s.mtime.size + 1)
        else:
            x = s.mtime - time_offset

        container = ax.errorbar(x, s.mvrad, s.msvrad, label=_label, 
                                picker=True, marker=next(markers), zorder=next(zorders), **kwargs)


        containers[inst] = list(container)

        if show_histogram:
            kw = dict(histtype='step', bins='doane', orientation='horizontal')
            hlabel = f'{s.mvrad.std():.2f} {self.units}'
            axh.hist(s.mvrad, **kw, label=hlabel)

        if show_masked:
            if versus_n:
                pass
            else:
                ax.errorbar(s.time[~s.mask] - time_offset, s.vrad[~s.mask], s.svrad[~s.mask],
                            **kwargs, color=container[0].get_color())

    if show_masked:
        if versus_n:
            pass
            # ax.errorbar(self.time[~self.mask] - time_offset, self.vrad[~self.mask], self.svrad[~self.mask],
            #             label='masked', fmt='x', ms=10, color='k', zorder=-2)
        else:
            ax.errorbar(self.time[~self.mask] - time_offset, self.vrad[~self.mask], self.svrad[~self.mask],
                        label='masked', fmt='x', ms=10, color='k', zorder=-2)

    if show_legend:
        leg = ax.legend()
        on_pick_legend = clickable_legend(fig, ax, leg)
        plt.connect('pick_event', on_pick_legend)

    if tooltips:
        from matplotlib.lines import Line2D
        from matplotlib.collections import LineCollection
        annotations = []
        axes_artists = []
        selected_inds = []
        selected_insts = []

        def _cleanup():
            for text in annotations:
                text.remove()
                annotations.remove(text)
            for art in axes_artists:
                art.remove()
                axes_artists.remove(art)
            for ind in selected_inds:
                selected_inds.remove(ind)

        def on_press(event):
            # print('press', event.key)
            if event.key in ('r',):
                for i, inst in zip(selected_inds, selected_insts):
                    i = self._index_from_instrument_index(i, inst)
                    self.remove_point(i)
                _cleanup()
                fig.canvas.draw_idle()
            if event.key == 'escape':
                _cleanup()
                kp_cid = None
                for k, _f in fig.canvas.callbacks.callbacks['key_press_event'].items():
                    if _f._obj == on_press:
                        kp_cid = k
                if kp_cid is not None:
                    fig.canvas.callbacks.disconnect(k)
                fig.canvas.draw_idle()

        def on_pick_point(event):
            _cleanup()            
            artist = event.artist
            if isinstance(artist, (Line2D, LineCollection)):
                # print(event.ind, artist)
                if isinstance(artist, Line2D):
                    ind = event.ind
                    matching_instrument = [k for k, v in containers.items() if artist in v]
                    if len(matching_instrument) == 0:
                        return
                    inst = matching_instrument[0]

                    _s = getattr(self, inst)

                    if event.ind.size > 1:
                        mint, maxt = _s.mtime[ind].min(), _s.mtime[ind].max()
                        miny, maxy = _s.mvrad[ind].min(), _s.mvrad[ind].max()
                        axins = ax.inset_axes([0.1, 0.5, 0.5, 0.4],
                            xlim=(mint - 0.1 * (maxt - mint), maxt + 0.1 * (maxt - mint)), 
                            ylim=(miny - 0.1 * (maxy - miny), maxy + 0.1 * (maxy - miny)), 
                            xticklabels=[], yticklabels=[]
                            )
                        axins.errorbar(_s.mtime[ind], _s.mvrad[ind], _s.msvrad[ind], fmt='o', ms=3,
                                       color=artist.get_color())
                        axins.margins(x=0.5)
                        axins.autoscale_view()
                        rectangle_patch, connector_lines = ax.indicate_inset_zoom(axins, edgecolor="black")
                        axes_artists.append(axins)
                        axes_artists.append(rectangle_patch)
                        for line in connector_lines:
                            axes_artists.append(line)
                    else:
                        ind = event.ind[0]
                        selected_inds.append(ind)
                        selected_insts.append(inst)
                        # print(_s.mtime[ind], _s.mvrad[ind], _s.msvrad[ind])

                        text = f'{inst} ({ind})\n'
                        text += f'{_s.mtime[ind]:9.5f}\n'
                        text += f'RV: {_s.mvrad[ind]:.1f} ± {_s.msvrad[ind]:.1f}'

                        annotations.append(
                            ax.annotate(text, (_s.mtime[ind], _s.mvrad[ind]), xycoords='data',
                                        xytext=(5, 10), textcoords='offset points', fontsize=9,
                                        bbox={'boxstyle': 'round', 'fc': 'w'}, arrowprops=dict(arrowstyle="-"))
                        )
                    # ax.annotate(f'{inst}', (0.5, 0.5), xycoords=artist, ha='center', va='center')
                    fig.canvas.draw()
                    # print(event.ind, artist.get_label())

                    _ = fig.canvas.mpl_connect('key_press_event', on_press)

        plt.connect('pick_event', on_pick_point)


    if show_histogram:
        axh.legend()

    ax.minorticks_on()

    ax.set_ylabel(f'RV [{self.units}]')
    if versus_n:
        ax.set_xlabel('observation')
    else:
        if remove_50000:
            ax.set_xlabel('BJD - 2450000 [days]')
        else:
            ax.set_xlabel('BJD - 2400000 [days]')

    if show_title:
        ax.set_title(self.star, loc='right')

    # from matplotlib.backend_tools import ToolBase, ToolToggleBase
    # tm = fig.canvas.manager.toolmanager
    # class InfoTool(ToolToggleBase):
    #     description = "Show extra information about each observation"
    #     image = os.path.abspath(os.path.join(os.path.dirname(__file__), 'data', 'info.svg'))
    #     # def enable(self, *args, **kwargs):
    #     #     self.figure.add_axes([1, 0, 0.3, 1])
    #     #     self.figure.canvas.draw_idle()
    # from PIL import UnidentifiedImageError
    # try:
    #     tm.add_tool("infotool", InfoTool)
    #     fig.canvas.manager.toolbar.add_tool(tm.get_tool("infotool"), "toolgroup")
    #     raise UnidentifiedImageError
    # except AttributeError:
    #     pass
    # except UnidentifiedImageError:
    #     pass


    if config.return_self:
        return self
    
    if show_histogram:
        return fig, (ax, axh)
    
    return fig, ax


# @plot_fast
def plot_quantity(self, quantity, ax=None, show_masked=False, instrument=None,
                  time_offset=0, remove_50000=False, tooltips=False, show_legend=True,
                  N_in_label=False, **kwargs):
    logger = setup_logger()
    if self.N == 0:
        if self.verbose:
            logger.error('no data to plot')
        return

    if not hasattr(self, quantity):
        logger.error(f"cannot find '{quantity}' attribute")
        return

    if ax is None:
        fig, ax = plt.subplots(1, 1, constrained_layout=True)
    else:
        fig = ax.figure

    kwargs.setdefault('marker', 'o')
    kwargs.setdefault('ls', '')
    kwargs.setdefault('capsize', 0)
    kwargs.setdefault('ms', 4)

    if remove_50000:
        time_offset = 50000

    instruments = self._check_instrument(instrument)

    for inst in instruments:
        s = self if self._child else getattr(self, inst)
        label = f'{inst:10s} ({s.N})' if N_in_label else inst
        if not self.only_latest_pipeline:
            i, p = label.split('_', 1)
            p = p.replace('_', '.')
            label = f'{i}-{p}'

        y = getattr(s, quantity).copy()
        try:
            ye = getattr(s, quantity + '_err').copy()
        except AttributeError:
            ye = np.zeros_like(y)

        if (nans := np.isnan(y)).any():
            if self.verbose:
                logger.warning(f'{nans.sum()} NaN values for {inst}')
            ye[nans] = np.nan

        if np.isnan(y).all() or np.isnan(ye).all():
            lines, *_ = ax.errorbar([], [], [],
                                    label=label, picker=True, **kwargs)
            continue

        lines, *_ = ax.errorbar(s.mtime - time_offset, y[s.mask], ye[s.mask],
                                label=label, picker=True, **kwargs)

        if show_masked:
            ax.errorbar(s.time[~s.mask] - time_offset, y[~s.mask], ye[~s.mask],
                        **kwargs, color=lines.get_color())

    if show_masked:
        ax.errorbar(self.time[~self.mask] - time_offset,
                    getattr(self, quantity)[~self.mask],
                    getattr(self, quantity + '_err')[~self.mask],
                    label='masked', fmt='x', ms=10, color='k', zorder=-2)

    if show_legend:
        leg = ax.legend()
        on_pick_legend = clickable_legend(fig, ax, leg)
        plt.connect('pick_event', on_pick_legend)

    ax.minorticks_on()

    delta = 'Δ' if self._did_adjust_means else ''

    ylabel = {
        quantity.lower(): quantity,
        'fwhm': f'{delta}FWHM [{self.units}]',
        'bispan': f'{delta}BIS [{self.units}]',
        'rhk': r"$\log$ R'$_{HK}$",
        'berv': 'BERV [km/s]',
    }

    ax.set_ylabel(ylabel[quantity.lower()])

    if remove_50000:
        ax.set_xlabel('BJD - 2450000 [days]')
    else:
        ax.set_xlabel('BJD - 2400000 [days]')

    if config.return_self:
        return self
    else:
        return fig, ax


plot_fwhm = partialmethod(plot_quantity, quantity='fwhm')
plot_bispan = partialmethod(plot_quantity, quantity='bispan')
plot_contrast = partialmethod(plot_quantity, quantity='contrast')
plot_rhk = partialmethod(plot_quantity, quantity='rhk')
plot_berv = partialmethod(plot_quantity, quantity='berv')


def plot_xy(self, x, y, ax=None, instrument=None, show_legend=True, **kwargs):
    logger = setup_logger()
    if self.N == 0:
        if self.verbose:
            logger.error('no data to plot')
        return

    if ax is None:
        fig, ax = plt.subplots(1, 1, constrained_layout=True)
    else:
        fig = ax.figure

    kwargs.setdefault('marker', 'o')
    kwargs.setdefault('ls', '')
    kwargs.setdefault('capsize', 0)
    kwargs.setdefault('ms', 4)

    instruments = self._check_instrument(instrument)

    for inst in instruments:
        s = self if self._child else getattr(self, inst)
        label = inst

        missing = False
        try:
            xdata = getattr(s, x).copy()
        except AttributeError:
            missing = True
        try:
            e_xdata = getattr(s, x + '_err').copy()
        except AttributeError:
            e_xdata = np.zeros_like(xdata)

        try:
            ydata = getattr(s, y).copy()
        except AttributeError:
            missing = True
        try:
            e_ydata = getattr(s, y + '_err').copy()
        except AttributeError:
            e_ydata = np.zeros_like(ydata)

        if missing:
            lines, *_ = ax.errorbar([], [], [],
                                    label=label, picker=True, **kwargs)
            continue

        ax.errorbar(xdata[s.mask], ydata[s.mask], e_xdata[s.mask], e_ydata[s.mask],
                    label=label, **kwargs)

    # if show_masked:
    #     ax.errorbar(self.time[~self.mask] - time_offset,
    #                 getattr(self, quantity)[~self.mask],
    #                 getattr(self, quantity + '_err')[~self.mask],
    #                 label='masked', fmt='x', ms=10, color='k', zorder=-2)

    if show_legend:
        leg = ax.legend()
        on_pick_legend = clickable_legend(fig, ax, leg)
        plt.connect('pick_event', on_pick_legend)

    ax.minorticks_on()

    delta = 'Δ' if self._did_adjust_means else ''

    # ylabel = {
    #     quantity.lower(): quantity,
    #     'fwhm': f'{delta}FWHM [{self.units}]',
    #     'bispan': f'{delta}BIS [{self.units}]',
    #     'rhk': r"$\log$ R'$_{HK}$",
    #     'berv': 'BERV [km/s]',
    # }

    # ax.set_ylabel(ylabel[quantity.lower()])

    if config.return_self:
        return self
    else:
        return fig, ax


# @plot_fast
def gls(self, ax=None, label=None, instrument=None, 
        fap=True, fap_method='baluev', adjust_means=config.adjust_means_gls,
        picker=True, **kwargs):
    """
    Calculate and plot the Generalised Lomb-Scargle periodogram of the radial
    velocities.

    Args:
        ax (matplotlib.axes.Axes):
            The matplotlib axes to plot on. If None, a new figure will be
            created.
        label (str):
            The label to use for the plot.
        instrument (str or list):
            Which instruments' data to include in the periodogram. Default is
            all instruments.
        fap (bool or float):
            Whether to show the false alarm probability. A value (not a
            percentage) can be provided to display a given FAP. Default is True.
        fap_method (str):
            Method used to estimate the FAP, passed directly to
            `astropy.timeseries.LombScargle`. Default is 'baluev'.
        adjust_means (bool):
            Whether to adjust (subtract) the weighted means of each instrument.
            Default is `config.adjust_means_gls`.
    """
    logger = setup_logger()
    if self.N == 0:
        if self.verbose:
            logger.error('no data to compute gls')
        return

    if not self._did_adjust_means and not adjust_means:
        logger.warning('gls() called before adjusting instrument means, '
                       'consider using the `adjust_means` argument')

    if instrument is not None:
        strict = kwargs.pop('strict', False)
        instrument = self._check_instrument(instrument, strict=strict, log=True)
        if instrument is None:
            return

        instrument_mask = np.isin(self.instrument_array, instrument)
        t = self.time[instrument_mask & self.mask].copy()
        y = self.vrad[instrument_mask & self.mask].copy()
        e = self.svrad[instrument_mask & self.mask].copy()
        if self.verbose:
            logger.info(f'calculating periodogram for instrument {instrument}')

        if adjust_means and not self._child:
            if self.verbose:
                logger.info('adjusting instrument means before gls')
            means = np.zeros_like(y)
            for i in instrument:
                mask = self.instrument_array[instrument_mask & self.mask] == i
                if len(y[mask]) > 0:
                    means = means + wmean(y[mask], e[mask]) * mask
            y = y - means

    else:
        t = self.time[self.mask].copy()
        y = self.vrad[self.mask].copy()
        e = self.svrad[self.mask].copy()

        if adjust_means and not self._child:
            if self.verbose:
                logger.info('adjusting instrument means before gls')
            means = np.zeros_like(y)
            for i in self.instruments:
                mask = self.instrument_array[self.mask] == i
                if len(y[mask]) > 0:
                    means = means + wmean(y[mask], e[mask]) * mask
            y = y - means

    self._gls = gls = LombScargle(t, y, e)

    maximum_frequency = kwargs.pop('maximum_frequency', 1.0)
    minimum_frequency = kwargs.pop('minimum_frequency', None)
    samples_per_peak = kwargs.pop('samples_per_peak', 10)
    kw = {
        'maximum_frequency': maximum_frequency,
        'minimum_frequency': minimum_frequency,
        'samples_per_peak': samples_per_peak
    }

    freq, power = gls.autopower(**kw)

    show_peak_fap = kwargs.pop('show_peak_fap', False)

    if ax is None:
        fig, ax = plt.subplots(1, 1, constrained_layout=True)
    else:
        fig = ax.figure

    if kwargs.pop('fill_between', False):
        kwargs.pop('lw', None)
        ax.fill_between(1/freq, 0, power, label=label, lw=0, **kwargs)
        ax.set_xscale('log')
    else:
        ax.semilogx(1/freq, power, picker=picker, label=label, **kwargs)

    if fap:
        fap_level = 0.01
        if isinstance(fap, float):
            fap_level = fap

        fap = gls.false_alarm_level(fap_level, method=fap_method, **kw)

        # if fap > 0.05 and fap_method == 'baluev':
        #     logger.warning('FAP is high (>5%), the analytical estimate may be underestimated. Using the bootstrap method instead.')
        #     fap = gls.false_alarm_level(fap_level, method='bootstrap')

        ax.axhline(fap, color='k', alpha=0.2, zorder=-1)

    if show_peak_fap:
        peak_per = 1/freq[np.argmax(power)]
        peak_power = np.max(power)
        peak_fap = gls.false_alarm_probability(peak_power, method=fap_method, **kw)
        ax.plot(peak_per, peak_power, 'o', color='r', zorder=1)
        ax.annotate(f'{peak_per:1.3f} days\nFAP: {peak_fap:1.1e}', (peak_per, peak_power), 
                    va='top', textcoords='offset points', xytext=(10, 0), zorder=1)

    ax.set(xlabel='Period [days]', ylabel='Normalized power', ylim=(0, None))
    ax.minorticks_on()

    if label is not None:
        ax.legend()

    if ax.get_legend() is not None:
        from matplotlib.text import Text
        leg = ax.get_legend()
        for text in leg.get_texts():
            text.set_picker(True)

        def on_pick_legend(event):
            handles, labels = ax.get_legend_handles_labels()
            artist = event.artist
            if isinstance(artist, Text):
                # print('handles:', handles)
                # print('labels:', labels)
                # print(artist.get_text())
                try:
                    h = handles[labels.index(artist.get_text())]
                    alpha_text = {None:0.2, 1.0: 0.2, 0.2:1.0}[artist.get_alpha()]
                    alpha_point = {None: 0.0, 1.0: 0.0, 0.2: 1.0}[artist.get_alpha()]
                    h.set_alpha(alpha_point)
                    artist.set_alpha(alpha_text)
                    fig.canvas.draw()
                except ValueError:
                    pass

        if 'pick_event' not in fig.canvas.callbacks.callbacks:
            plt.connect('pick_event', on_pick_legend)


    if config.return_self:
        return self
    else:
        return fig, ax


# @plot_fast
def gls_quantity(self, quantity, ax=None, instrument=None,
                 fap=True, fap_method='baluev',
                 adjust_means=True, picker=True, **kwargs):
    """
    Calculate and plot the Generalised Lomb-Scargle periodogram of the given
    quantity (e.g. fwhm, rhk, etc.)

    Args:
        quantity (str):
            The quantity to calculate for which to compute the periodogram.
        ax (matplotlib.axes.Axes):
            The matplotlib axes to plot on. If None, a new figure will be
            created.
        instrument (str or list):
            Which instruments' data to include in the periodogram. Default is
            all instruments.
        fap (bool or float):
            Whether to show the false alarm probability. A value (not a
            percentage) can be provided to display a given FAP. Default is True.
        fap_method (str):
            Method used to estimate the FAP, passed directly to
            `astropy.timeseries.LombScargle`. Default is 'baluev'.
        adjust_means (bool):
            Whether to adjust (subtract) the weighted means of each instrument.
            Default is `config.adjust_means_gls`.
    """
    logger = setup_logger()
    if not hasattr(self, quantity):
        if self.verbose:
            logger.error(f"cannot find '{quantity}' attribute")
        return

    if self.N == 0:
        if self.verbose:
            logger.error('no data to compute gls')
        return

    if not self._did_adjust_means and not adjust_means:
        logger.warning('gls() called before adjusting instrument means, '
                       'consider using the `adjust_means` argument')

    strict = kwargs.pop('strict', False)
    instrument = self._check_instrument(instrument, strict=strict, log=True)
    if instrument is None:
        return

    instrument_mask = np.isin(self.instrument_array, instrument)
    final_mask = instrument_mask & self.mask

    t = self.time[final_mask].copy()
    y = getattr(self, quantity)[final_mask].copy()
    try:
        ye = getattr(self, quantity + '_err')[final_mask].copy()
    except AttributeError:
        ye = None

    if self.verbose:
        logger.info(f'calculating periodogram for instrument {instrument}')

    nan_mask = np.isnan(y)
    if nan_mask.any():
        if self.verbose:
            logger.warning(f'{quantity} contains NaNs, ignoring them')

    if adjust_means and not self._child:
        if self.verbose:
            logger.info('adjusting instrument means before gls')
        means = np.zeros_like(y)
        for i in instrument:
            mask = (self.instrument_array[final_mask] == i) & ~nan_mask
            if len(y[mask]) > 0:
                # print(wmean(y[mask], ye[mask]))
                means += wmean(y[mask], ye[mask]) * mask
        y = y - means

    t = t[~nan_mask]
    y = y[~nan_mask]
    ye = ye[~nan_mask]

    if len(y) == 0:
        logger.error('no data left to compute gls after removing NaNs')
        return

    if ax is None:
        fig, ax = plt.subplots(1, 1, constrained_layout=True)
    else:
        fig = ax.figure

    maximum_frequency = kwargs.pop('maximum_frequency', 1.0)
    minimum_frequency = kwargs.pop('minimum_frequency', None)
    samples_per_peak = kwargs.pop('samples_per_peak', 10)
    kw = {
        'maximum_frequency': maximum_frequency,
        'minimum_frequency': minimum_frequency,
        'samples_per_peak': samples_per_peak
    }

    gls = LombScargle(t, y, ye)
    freq, power = gls.autopower(**kw)

    ax.semilogx(1/freq, power, picker=picker, **kwargs)

    if fap:
        fap_level = 0.01
        if isinstance(fap, float):
            fap_level = fap

        fap = gls.false_alarm_level(fap_level, method=fap_method, **kw)

        # if fap > 0.05 and fap_method == 'baluev':
        #     logger.warning('FAP is high (>5%), the analytical estimate may be underestimated. Using the bootstrap method instead.')
        #     fap = gls.false_alarm_level(fap_level, method='bootstrap', **kw)

        ax.axhline(fap, color='k', alpha=0.2, zorder=-1)

    ax.set(xlabel='Period [days]', ylabel='Normalized power', ylim=(0, None))
    ax.minorticks_on()

    if config.return_self:
        return self
    else:
        return fig, ax


gls_fwhm = partialmethod(gls_quantity, quantity='fwhm')
gls_bispan = partialmethod(gls_quantity, quantity='bispan')
gls_rhk = partialmethod(gls_quantity, quantity='rhk')



def window_function(self, ax1=None, ax2=None, instrument=None, crosshair=False, **kwargs):
    """
    Calculate and plot the window function of the observed times.

    Args:
        ax1 (matplotlib.axes.Axes):
            An axes to plot the window function vs period. If None, a new figure
            will be created.
        ax2 (matplotlib.axes.Axes):
            An axes to plot the periodogram vs frequency. If None, a new figure
            will be created.
        instrument (str or list):
            Which instruments' data to include in the window function.
        crosshair (bool):
            If True, a crosshair will be drawn on the plot.
    """
    logger = setup_logger()

    if self.N == 0:
        if self.verbose:
            logger.error('no data to compute window function')
        return

    if ax1 is None:
        fig, (ax1, ax2) = plt.subplots(2, 1, constrained_layout=True)
    else:
        fig = ax1.figure

    if instrument is not None:
        strict = kwargs.pop('strict', False)
        instrument = self._check_instrument(instrument, strict=strict)
        if instrument is not None:
            instrument_mask = np.isin(self.instrument_array, instrument)
            t = self.time[instrument_mask & self.mask]
            ye = self.svrad[instrument_mask & self.mask]
            if self.verbose:
                logger.info(f'calculating window function for instrument {instrument}')
    else:
        t = self.time[self.mask]
        ye = self.svrad[self.mask]

    wf = LombScargle(t, np.ones_like(t), ye / np.std(ye),
                     fit_mean=False, center_data=False)

    freq, power = wf.autopower(maximum_frequency=1.1, 
                               samples_per_peak=20, method='cython')
    ax1.semilogx(1/freq, power, **kwargs)
    ax1.set(xlabel='Period [days]', ylabel='Window function')

    ax2.plot(freq, power, **kwargs)
    ax2.set(xlabel='Frequency [1/day]', ylabel='Window function')

    for x in (365.25, 1.0, 1 - 1.0/365.25):
        ax1.axvline(x, color='k', alpha=0.2, zorder=-1)
        ax2.axvline(1/x, color='k', alpha=0.2, zorder=-1)
    
    if crosshair:
        blitted_cursor = BlittedCursor((ax1, ax2), horizontal=False,
                                    transforms_x=(lambda x:x, lambda x:1/x))
        fig.canvas.mpl_connect('motion_notify_event', blitted_cursor.on_mouse_move)
        return fig, (ax1, ax2), blitted_cursor
        # from matplotlib.widgets import MultiCursor
        # cursor = MultiCursor(fig.canvas, (ax1, ax2), color='r',
        #                      lw=0.5, horizOn=False, vertOn=True)
        # return fig, (ax1, ax2), (cursor)
    else:
        return fig, (ax1, ax2)

def histogram_svrad(self, ax=None, instrument=None, label=None):
    """ Plot an histogram of the radial velocity uncertainties. 
    
    Args:
        ax (matplotlib.axes.Axes):
            The matplotlib axes to plot on. If None, a new figure will be
            created.
        instrument (str or list):
            Which instruments' data to include in the histogram.
        label (str):
            The label to use for the plot.
    """
    if ax is None:
        fig, ax = plt.subplots(1, 1, constrained_layout=True)
    else:
        fig = ax.figure

    instruments = self._check_instrument(instrument)

    for inst in instruments:
        s = self if self._child else getattr(self, inst)

        if label is None:
            _label = inst
            if not self.only_latest_pipeline:
                i, p = _label.split('_', 1)
                p = p.replace('_', '.')
                _label = f'{i}-{p}'
        else:
            _label = label

        kw = dict(bins=40, histtype='step', density=False, lw=2)
        ax.hist(s.msvrad, label=_label, **kw)
    ax.legend()
    ax.set(xlabel=f'RV uncertainty [m/s]', ylabel='Number')




import os
import numpy as np
import matplotlib.pyplot as plt

from arvi.headers import get_headers
from barycorrpy import get_BC_vel
from astropy.coordinates import SkyCoord
from astropy.time import Time
from astropy import units as u
from astropy import constants as const
from astropy.timeseries import LombScargle
from tqdm import tqdm

from .setup_logger import logger
from .config import config


def correct_rvs(self, simple=False, H=None, save_files=False, plot=True):
    """
    """
    import pickle

    if hasattr(self, '_did_correct_berv') and self._did_correct_berv:
        logger.info('Already corrected for the BERV! Not doing anything.')
        return

    path = os.path.dirname(__file__)
    path = os.path.join(path, 'data')
    pkl = os.path.join(path, 'berv_espresso_sine.pkl')
    berv_espresso = pickle.load(open(pkl, 'rb'))
    
    if simple:
        logger.info('Correcting RVs with a previously-fitted sinusoid function')
        _f = berv_espresso['func'].replace('lambda t: ', '')
        logger.info(f': {_f}')
        f = eval(berv_espresso['func'])
        if plot:
            _, ax = self.plot()
            ax.plot(self._tt, f(self._tt) + self.vrad.mean(), 'k')
            _, axgls = self.gls(label='before')
        
        self.vrad = self.vrad + f(self.time)

        if plot:
            self.gls(ax=axgls, label='after')

        return f(self.time)

    else:
        logger.info('Correcting RVs with actual difference between BERVs')
        logger.info('(basically, use BERV_barycorrpy for BERV correction)')

        old_vrad = self.vrad.copy()

        _, berv = BERV(self, H=H, use_gaia_meassurements=True, plx=self.gaia.plx, 
                       plot=False, ignore_mask=True)

        if plot:
            fig, axs = plt.subplots(2, 1, constrained_layout=True, height_ratios=(3, 1), sharex=True)
            _, ax = self.plot(ax=axs[0])
            _, axgls = self.gls(label='before')

        # undo secular acceleration, if it was done
        _did_secular_acceleration = self._did_secular_acceleration
        self._undo_secular_acceleration()

        # transform RVs: RV --> RV - BERVpipe + BERVbarycorrpy

        diff = berv[self.star]['berv_barycorrpy'] - berv[self.star]['berv_pipeline']

        if save_files:
            i_inst = np.hstack([np.arange(n) for n in self.NN.values()])
            with open(f'{self.star}_berv_correction.rdb', 'w') as rdb:
                rdb.write('# time\n')
                rdb.write('# vrad\n')
                rdb.write('# svrad\n')
                rdb.write('# berv - BERV value from header\n')
                rdb.write('# berv_pipe - BERV from header corrected for 1.55e-8 factor\n')
                rdb.write('# berv_barycorrpy - BERV value from barycorrpy\n')
                rdb.write('# diff - difference between berv_barycorrpy and berv_pipe\n')
                rdb.write('# vrad_berv_corrected = vrad + diff\n')
                rdb.write('# instrument\n')
                rdb.write('# i - index\n')
                rdb.write('# i_inst - index within the instrument\n')
                rdb.write('#\n')
                rdb.write('# --> TO CORRECT vrad, we ** add the diff column **\n')
                rdb.write('# --> the result of this operation is in column vrad_berv_corrected\n')
                rdb.write('# --> vrad_berv_corrected is already corrected for the secular acceleration, vrad is not\n')
                rdb.write('#\n')
                # 
                cols = [
                    'time', 'vrad', 'svrad', 
                    'berv', 'berv_pipe', 'berv_barycorrpy', 'diff', 'vrad_berv_corrected',
                    'instrument', 'i', 'i_inst'
                ]
                rdb.write('# ' + '\t'.join(cols) + '\n')
                for i, t in enumerate(self.time):
                    rdb.write(f'{t:11.5f}\t')
                    # if _did_secular_acceleration:
                    #     rdb.write(f'{old_vrad[i]:13.5f}\t')
                    # else:
                    rdb.write(f'{self.vrad[i]:13.7f}\t')
                    rdb.write(f'{self.svrad[i]:13.7f}\t')
                    rdb.write(f'{self.berv[i]:15.7f}\t')
                    rdb.write(f'{berv[self.star]["berv_pipeline"][i]/1e3:15.7f}\t')
                    rdb.write(f'{berv[self.star]["berv_barycorrpy"][i]/1e3:15.7f}\t')
                    rdb.write(f'{diff[i]:15.7f}\t')
                    rdb.write(f'{self.vrad[i] + diff[i]:13.7f}\t')
                    rdb.write(f'{self.instrument_array[i]}\t')
                    rdb.write(f'{i}\t')
                    rdb.write(f'{i_inst[i]}\t')
                    rdb.write('\n')

        self.add_to_vrad(diff)
        self._did_correct_berv = True
        self._did_secular_acceleration = True # "automatically", by using BERV_barycorrpy
        self._did_secular_acceleration_simbad = False
        self._did_secular_acceleration_epoch = Time('J2016').jd - 24e5

        # the secular acceleration hadn't been done, but it was introduced by
        # BERV_barycorrpy, so we need to undo it
        if not _did_secular_acceleration:
            self._undo_secular_acceleration()

        if plot:
            self.plot(ax=axs[0], marker='+', ms=5)
            axs[1].plot(self.time, old_vrad - self.vrad, '.k', label='old RV - new RV')
            ma = np.abs(axs[1].get_ylim()).max()
            axs[1].set(ylim=(-ma, ma), xlabel=axs[0].get_xlabel(), ylabel='RV difference [m/s]')
            self.gls(ax=axgls, label='after')

        return diff

def get_A_and_V_from_lesta(self, username=config.username):
    try:
        import paramiko
    except ImportError:
        raise ImportError("paramiko is not installed. Please install it with 'pip install paramiko'")

    logs = []
    for f in self.raw_file:
        f = f.replace('espresso/', '/projects/astro/ESPRESSODRS/')
        f = f.replace('nirps/', '/projects/astro/NIRPSDRS/')
        f = f.replace('.fits', '_SCIENCE_FP.log')
        f = f.replace('reduced', 'log')
        f = f.replace('r.ESPRE', 'ESPRESSO')
        logs.append(f)

    A, V = [], []

    try:
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect("lesta02.astro.unige.ch", username=username, timeout=5)
    except Exception as e:
        if 'getaddrinfo failed' in str(e):
            jump = paramiko.SSHClient()
            jump.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            jump.connect('login01.astro.unige.ch', username=username, timeout=5)
            jump_transport = jump.get_transport()
            jump_channel = jump_transport.open_channel('direct-tcpip', ('10.194.64.162', 22), ('129.194.64.20', 22))
            ssh = paramiko.SSHClient()
            ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            ssh.connect('lesta02.astro.unige.ch', username=username, sock=jump_channel)
        else:
            raise e

    with ssh.open_sftp() as sftp:
        pbar = tqdm(logs, total=len(logs), unit='file', desc='Reading logs')
        for f in pbar:
            with sftp.open(f) as fp:
                pattern1 = 'Sun-Earth'
                pattern2 = "Barycentric Observer's Velocity"
                for line in fp:
                    if pattern1 in line:
                        value = line.strip().split(':')[-1].replace('\x1b[32m', '').replace('\x1b[0m', '').replace(' ', '')
                        A.append(float(value))
                    if pattern2 in line:
                        value = line.strip().split(':')[-1].replace('\x1b[32m', '').replace('\x1b[0m', '').replace(' ', '')
                        V.append(float(value))

    ssh.close()

    return np.array(A), np.array(V)


def BERV(self, H=None, use_gaia_meassurements=False, plx=None,
         A=None, V=None, plot=True, ignore_mask=False, verbose=False, dpi=None):
    """ Calculate Barycentric Radial Velocity with barycorr and compare with pipeline

    Args:
        H (list, optional):
            List of (CCF/S1D/etc) headers for the target. If None, try to
            download the CCF files to get the headers.
        use_gaia_meassurements (bool, optional):
            Use Gaia coordinates and proper motions instead of those in the headers.
        plx (float, optional):
            Value of stellar parallax [mas] to use in barycorr.
        A (array, optional):
            Earth-Sun distance [AU] for each BJD (found in the pipeline logs).
        V (array, optional):
            Earth's orbital velocity [km/s] for each BJD (found in the pipeline logs).
        plot (bool, optional):
            Plot the results.
    """
    if H is None:
        H = get_headers(self, check_lesta=False, check_exo2=False, instrument='ESPRE')

    if len(H) != self.N:
        raise ValueError(f'Expected {self.N} headers (in `H`), got {len(H)}')

    if 'HARPS' in H[0]['INSTRUME'] or 'NIRPS' in H[0]['INSTRUME']:
        obsname = 'lasilla'
    elif 'ESPRESSO' in H[0]['INSTRUME']:
        obsname = 'paranal'
    else:
        raise ValueError('unknown instrument')

    bjd = np.array([h['HIERARCH ESO QC BJD'] for h in H])
    bjd -= 24e5
    berv_pipeline = np.array([h['HIERARCH ESO QC BERV'] for h in H])

    # in the pipeline, the BERV is used to shift wavelenghts with this formula
    # berv_factor = (1 + 1.55e-8) * (1 + BERV/c)
    # The 1.55e-8 factor is an average of some relativistic effects, which are
    # probably already included in the BERV calculated from barycorrpy.
    # Therefore, we compute an "effective" BERV from the pipeline doing
    # (1 + 1.55e-8) * (1 + BERV/c) = 1 + effBERV/c 
    # => effBERV = ((1 + 1.55e-8) * (1 + BERV/c) - 1) * c

    if A is None and V is None:
        if verbose:
            logger.info("Using mean value for Earth-Sun distance and Earth's orbital velocity")

    if A is None:
        Φobs = const.G * const.M_sun / const.au + const.G * const.M_earth / const.R_earth
    else:
        A = np.atleast_1d(A) * u.km
        Φobs = const.G * const.M_sun / A + const.G * const.M_earth / const.R_earth

    if V is None:
        V = 29785 *u.m / u.second
    else:
        V = np.atleast_1d(V) * u.km / u.second

    f = 1 / (1 - Φobs / const.c**2 - V**2 / (2*const.c**2))
    c = const.c.to(u.km / u.second).value
    berv_pipeline = (f * (1 + berv_pipeline/c) - 1) * c


    tmmean = np.array([h['HIERARCH ESO QC TMMEAN USED'] for h in H])
    mjdobs = np.array([h['MJD-OBS'] for h in H])
    texp = np.array([h['EXPTIME'] for h in H])
    jd = mjdobs + 24e5 + 0.5 + (texp * tmmean)/60/60/24

    if verbose:
        logger.info(f"Unique exposure times: {np.unique(texp)}")

    berv = []
    if verbose:
        pbar = enumerate(jd)
    else:
        pbar = tqdm(enumerate(jd), total=len(jd),
                    unit='observation', desc='Computing BERV')

    for i, _jd in pbar:
        if use_gaia_meassurements:
            if not hasattr(self, 'gaia'):
                raise ValueError('No Gaia data available')
            
            target = self.gaia.coords
            pmra = self.gaia.pmra
            pmdec = self.gaia.pmdec
            epoch = Time('J2016').jd
        else:
            ra = H[i]['* TARG ALPHA'][0]
            ra = f'{ra:09.2f}'
            ra = ra[:2] + 'h' + ra[2:4] + 'm' + ra[4:] + 's'

            dec = H[i]['* TARG DELTA'][0]
            if dec < 0:
                dec = f'{dec:010.2f}'
            else:
                dec = f'{dec:09.2f}'
            if dec.startswith('-'):
                dec = dec[:3] + 'd' + dec[3:5] + 'm' + dec[5:] + 's'
            else:
                dec = dec[:2] + 'd' + dec[2:4] + 'm' + dec[4:] + 's'

            target = SkyCoord(ra, dec)
            pmra = H[i]['* TARG PMA'][0] * 1e3
            pmdec = H[i]['* TARG PMD'][0] * 1e3
            epoch = Time('J2000').jd

        if verbose:
            logger.info(f'jd: {_jd}')
            logger.info(f'\t ra: {target.ra}')
            logger.info(f'\t dec: {target.dec}')
            logger.info(f'\t pmra: {pmra}')
            logger.info(f'\t pmdec: {pmdec}')
                        

        px = plx or 0.0
        out = get_BC_vel(_jd, obsname=obsname, rv=0.0, px=px, zmeas=0.0, epoch=epoch,
                         ra=target.ra.value, dec=target.dec.value, pmra=pmra, pmdec=pmdec)
        # print(out[1][3])
        berv.append(out[0])
    
    berv = np.array(berv).flatten()

    if ignore_mask: # ignore the system's masked points
        pass
    else: # mask points in the BERV output as well
        bjd = bjd[self.mask]
        berv = berv[self.mask]
        berv_pipeline = berv_pipeline[self.mask]

    fig = None
    if plot:
        fig, axs = plt.subplots(2, 1, figsize=(8, 6), dpi=dpi, sharex=True,
                                constrained_layout=True)
    
        axs[0].set_title(f'{self.star}', loc='right')
        axs[0].plot(bjd, berv_pipeline*1e3, '.', label='pipeline', alpha=0.5)
        axs[0].plot(bjd, berv, '.', label='barycorrpy', alpha=0.5)
        axs[0].legend(bbox_to_anchor=(0.0, 1.15), loc=2, borderaxespad=0., ncol=2)
        axs[0].set(xlabel='BJD - 2450000', ylabel='BERV [m/s]')


        if plx is not None:
            epoch = 55500
            sa = self.secular_acceleration(just_compute=True)
            print('sa:', sa)
            sec_acc = sa.value * (bjd - epoch) / 365.25

            axs[0].plot(bjd, sec_acc)

            # fitp = np.polyfit(bjd - epoch, diff, 1)
            # axs[1].plot(bjd, np.polyval(fitp, bjd - epoch))
            # axs[1].plot(bjd, np.mean(diff) + diff - np.polyval(fitp, bjd - epoch), '.')

        if plx is None:
            diff = berv - berv_pipeline*1e3
            label=r'BERV$_{\rm barycorrpy}$ - BERV$_{\rm pipeline}$'
        else:
            diff = berv + sec_acc - berv_pipeline*1e3
            label=r'BERV$_{\rm barycorrpy}$ (+SA) - BERV$_{\rm pipeline}$'

        axs[1].plot(bjd, diff, 'k.', label=label)
        axs[1].axhline(np.mean(diff), ls='--', c='k', alpha=0.1)

        text = axs[1].text(bjd.max(), diff.min() + 0.1*diff.ptp(), 
                           f'ptp: {diff.ptp()*1e2:.2f} cm/s',
                           ha='right', va='bottom', color='g', alpha=0.8)
        axs[1].plot([bjd[np.argmax(diff)], bjd.max() + 0.05 * bjd.ptp()], 
                    [np.max(diff), np.max(diff)], 'g--', alpha=0.3)
        axs[1].plot([bjd[np.argmin(diff)], bjd.max() + 0.05 * bjd.ptp()], 
                    [np.min(diff), np.min(diff)], 'g--', alpha=0.3)

        ax = axs[1].twinx()
        diff_cms = 1e2*(diff - np.mean(diff))
        ax.plot(bjd, diff_cms, alpha=0)
        ma = np.max(np.abs(ax.get_ylim()))
        ax.set_ylim(-1 - 5*round(ma/5), 1 + 5*round(ma/5))
        ax.set(ylabel='diff - mean(diff) [cm/s]')
        axs[1].set_ylim(np.mean(diff)-ma/100, np.mean(diff)+ma/100)

        axs[1].legend(bbox_to_anchor=(0.0, 1.15), loc=2, borderaxespad=0.)
        axs[1].set(xlabel='BJD - 2450000', ylabel='diff [m/s]')
    
        # adjust_text([text], va='bottom')

    return fig, {
        self.star: {
            'bjd': bjd,
            'berv_pipeline': berv_pipeline*1e3,
            'berv_barycorrpy': berv
        }
    }


def plot_BERV_correction(self, H, A, V, berv2=None, berv6=None,
                         inset=False, inset_range=(3, 5)):
    fig, axs = plt.subplot_mosaic('ab\ncc\ndd\nee', constrained_layout=True, figsize=(2*3.57, 10))

    if berv2 is None:
        _, berv2 = BERV(self, H, plot=False)
    if berv6 is None:
        _, berv6 = BERV(self, H, A=A, V=V, plot=False)

    self.plot(ax=axs['a'], ms=2)
    axs['a'].set_title('original', loc='right')
    self.gls(ax=axs['e'], label='original', color='r', alpha=0.5,
             fill_between=True, samples_per_peak=20)

    temp_vrad = self.vrad.copy()
    self.vrad[self.mask] = self.vrad[self.mask] - berv2[self.star]['berv_pipeline'].value + berv6[self.star]['berv_pipeline'].value

    self.plot(ax=axs['b'], ms=2)
    axs['b'].set_title('after correction', loc='right')

    diff = temp_vrad[self.mask] - self.vrad[self.mask]

    axs['c'].plot(self.mtime, diff, 'k.')
    axs['c'].set_title('RV difference', loc='right')
    axs['c'].set(xlabel='BJD - 2450000', ylabel='RV diff [m/s]')

    text = axs['c'].text(self.mtime.max(), diff.min() + 0.1*diff.ptp(), 
                         f'ptp: {diff.ptp()*1e2:.2f} cm/s',
                         ha='right', va='bottom', color='g', alpha=0.8)
    axs['c'].plot([self.mtime[np.argmax(diff)], self.mtime.max() + 0.05 * self.mtime.ptp()], 
                [np.max(diff), np.max(diff)], 'g--', alpha=0.3)
    axs['c'].plot([self.mtime[np.argmin(diff)], self.mtime.max() + 0.05 * self.mtime.ptp()], 
                [np.min(diff), np.min(diff)], 'g--', alpha=0.3)


    f, p = LombScargle(self.mtime, diff).autopower(maximum_frequency=1.0, samples_per_peak=10)
    axs['d'].semilogx(1/f, p, color='k', alpha=0.8)
    axs['d'].vlines([365.25, 365.25/2], 0, 1, color='k', ls='--', alpha=0.3)
    axs['d'].set(xlabel='Period [days]', ylabel='normalized power', ylim=(0, 1))
    axs['d'].set_title('GLS of RV difference', loc='right')

    if inset:
        inset = axs['d'].inset_axes(bounds=[0.15, 0.3, 0.3, 0.6])
        m = (1/f > inset_range[0]) & (1/f < inset_range[1])
        inset.plot(1/f[m], p[m], color='k', alpha=0.8)
        inset.set(xlim=inset_range, yticks=[])
        inset.minorticks_on()

    self.gls(ax=axs['e'], label='after correction', color='g', alpha=1,
             lw=0.8, samples_per_peak=20)
    axs['e'].set(xlabel='Period [days]', ylabel='normalized power')
    axs['e'].sharex(axs['d'])

    self.vrad = temp_vrad
    return fig


import os
from io import StringIO
from csv import DictReader
import requests

DATA_PATH = os.path.dirname(__file__)
DATA_PATH = os.path.join(DATA_PATH, 'data')

QUERY = """
SELECT TOP 20 gaia_source.designation, gaia_source.source_id,
              gaia_source.ra, gaia_source.dec, 
              gaia_source.parallax, gaia_source.pmra, gaia_source.pmdec,
              gaia_source.ruwe, gaia_source.phot_g_mean_mag, gaia_source.bp_rp, 
              gaia_source.radial_velocity, gaia_source.radial_velocity_error
FROM gaiadr3.gaia_source 
WHERE 
CONTAINS(
	POINT('ICRS',gaiadr3.gaia_source.ra,gaiadr3.gaia_source.dec),
	CIRCLE(
		'ICRS',
		COORD1(EPOCH_PROP_POS({ra},{dec},{plx},{pmra},{pmdec},{rv},2000,2016.0)),
		COORD2(EPOCH_PROP_POS({ra},{dec},{plx},{pmra},{pmdec},{rv},2000,2016.0)),
		0.001388888888888889)
)=1
"""

QUERY_ID = """
SELECT TOP 20 gaia_source.designation, gaia_source.source_id,
              gaia_source.ra, gaia_source.dec, 
              gaia_source.parallax, gaia_source.pmra, gaia_source.pmdec,
              gaia_source.ruwe, gaia_source.phot_g_mean_mag, gaia_source.bp_rp, 
              gaia_source.radial_velocity, gaia_source.radial_velocity_error
FROM gaiadr3.gaia_source 
WHERE 
gaia_source.source_id = {id}
"""

translate = {
    'Proxima': '5853498713190525696',
    'GJ699': '4472832130942575872',
    'LS II +14 13': '4318465066420528000',
}


def run_query(query):
    url = 'https://gea.esac.esa.int/tap-server/tap/sync'
    data = dict(query=query, request='doQuery', lang='ADQL', format='csv')
    try:
        response = requests.post(url, data=data, timeout=2)
    except requests.ReadTimeout as err:
        raise IndexError(err)
    except requests.ConnectionError as err:
        raise IndexError(err)
    return response.content.decode()

def parse_csv(csv):
    reader = DictReader(StringIO(csv))
    return list(reader)


class gaia:
    """
    A very simple wrapper around a TAP query to gaia for a given target. This
    class simply runs a few TAP queries and stores the result as attributes.

    Attributes:
        ra (float): right ascension
        dec (float): declination
        coords (SkyCoord): coordinates as a SkyCoord object
        dr3_id (int): Gaia DR3 identifier
        plx (float): parallax
        radial_velocity (float): radial velocity
    """
    def __init__(self, star:str, simbad=None, _debug=False):
        """
        Args:
            star (str): The name of the star to query simbad
        """
        from astropy.coordinates import SkyCoord

        self.star = star

        if simbad is None:
            from .simbad_wrapper import simbad as Simbad
            simbad = Simbad(star)
            if _debug:
                print(simbad)

        ra = simbad.ra
        dec = simbad.dec
        plx = simbad.plx
        pmra = simbad.pmra
        pmdec = simbad.pmdec
        rv = simbad.rvz_radvel
        args = dict(ra=ra, dec=dec, plx=plx, pmra=pmra, pmdec=pmdec, rv=rv) 

        try:
            if star in translate:
                table = run_query(query=QUERY_ID.format(id=translate[star]))
            elif hasattr(simbad, 'gaia_id') and simbad.gaia_id is not None:
                table = run_query(query=QUERY_ID.format(id=simbad.gaia_id))
            else:
                table = run_query(query=QUERY.format(**args))
            
            if _debug:
                print('table:', table)

            results = parse_csv(table)[0]
        except IndexError:
            raise ValueError(f'Gaia query for {star} failed')
        
        try:
            self.dr3_id = int(results['source_id'])
        except KeyError:
            raise ValueError(f'Gaia query for {star} failed')

        self.ra = float(results['ra'])
        self.dec = float(results['dec'])
        self.pmra = float(results['pmra'])
        self.pmdec = float(results['pmdec'])
        self.coords = SkyCoord(self.ra, self.dec, unit='deg')
        self.plx = float(results['parallax'])
        try:
            self.radial_velocity = float(results['radial_velocity'])
        except ValueError:
            self.radial_velocity = None
        try:
            self.radial_velocity_error = float(results['radial_velocity_error'])
        except ValueError:
            self.radial_velocity_error = None

        return

    def __repr__(self):
        return f'{self.star} (DR3 id={self.dr3_id})'



import sys

def setup_logger():
    from loguru import logger
    try:
        import marimo as mo
        if mo.running_in_notebook():
            raise NotImplementedError
    except (NotImplementedError, AttributeError):
        pass
    except (ImportError, ModuleNotFoundError):
        logger.remove()
    else:
        logger.remove()

    logger.configure(extra={"indent": ""})
    logger.add(
        sys.stdout,
        colorize=True,
        # format="<green>{time:YYYY-MM-DDTHH:mm:ss}</green> <level>{message}</level>",
        format="{extra[indent]}<level>{message}</level>",
    )

    return logger



import re

STARS = {
    '_51Peg': '51 Peg',
    'Barnard': 'GJ699',
    "Barnard's": 'GJ699',
    'Ross128': 'Ross  128',
    'Ross 128': 'Ross  128',
    # 
    'Teegarden': 'GAT1370',
    "Teegarden's Star": 'GAT1370',
    # 
    "Smethells 20": 'TIC464410508',
}


def translate(star, ngc=False, ic=False):
    # known translations
    if star in STARS:
        return STARS[star]

    # regex translations
    if ngc:
        NGC_match = re.match(r'NGC([\s\d]+)No([\s\d]+)', star)
        if NGC_match:
            cluster = NGC_match.group(1).replace(' ', '')
            target = NGC_match.group(2).replace(' ', '')
            return f'Cl* NGC {cluster} MMU {target}'
    if ic:
        IC_match = re.match(r'IC([\s\d]+)No([\s\d]+)', star)
        if IC_match:
            cluster = IC_match.group(1).replace(' ', '')
            target = IC_match.group(2).replace(' ', '')
            return f'Cl* IC {cluster} MMU {target}'

    return star



import os
import numpy as np

from .setup_logger import logger

try:
    import kima
    from kima.pykima.utils import chdir
    from kima import distributions
    from kima import RVData, RVmodel
    kima_available = True
except ImportError:
    kima_available = False


def try_to_guess_prior(model, prior):
    if 'jitter' in prior:
        return 'Jprior'
    if 'vsys' in prior:
        return 'Cprior'
    return None


def run_kima(self, run=False, load=False, run_directory=None, priors={}, **kwargs):
    if not kima_available:
        raise ImportError('kima not available, please install with `pip install kima`')

    instruments = [inst for inst in self.instruments if self.NN[inst] > 1]
    time = [getattr(self, inst).mtime for inst in instruments]
    vrad = [getattr(self, inst).mvrad for inst in instruments]
    err = [getattr(self, inst).msvrad for inst in instruments]
    data = RVData(time, vrad, err, instruments=instruments)

    fix = kwargs.pop('fix', False)
    npmax = kwargs.pop('npmax', 1)
    model = RVmodel(fix=fix, npmax=npmax, data=data)

    model.trend = kwargs.pop('trend', False)
    model.degree = kwargs.pop('degree', 0)

    model.studentt = kwargs.pop('studentt', False)
    model.enforce_stability = kwargs.pop('enforce_stability', False)
    model.star_mass = kwargs.pop('star_mass', 1.0)

    if kwargs.pop('gaussian_priors_individual_offsets', False):
        from kima.pykima.utils import get_gaussian_priors_individual_offsets
        model.individual_offset_prior = get_gaussian_priors_individual_offsets(data, use_std=True)

    if kwargs.pop('kuma', False):
        model.conditional.eprior = distributions.Kumaraswamy(0.867, 3.03)

    for k, v in priors.items():
        try:
            if 'conditional' in k:
                setattr(model.conditional, k.replace('conditional.', ''), v)
            else:
                setattr(model, k, v)

        except AttributeError:
            msg = f'`RVmodel` has no attribute `{k}`, '
            if guess := try_to_guess_prior(model, k):
                msg += f'did you mean `{guess}`?'
            logger.warning(msg)
            return

    if run_directory is None:
        run_directory = os.getcwd()

    if run:
        
        # TODO: use signature of kima.run to pop the correct kwargs
        # model_name = model.__class__.__name__
        # model_name = f'kima.{model_name}.{model_name}'
        # signature, defaults = [sig for sig in kima.run.__nb_signature__ if model_name in sig[0]]

        with chdir(run_directory):
            kima.run(model, **kwargs)
        
    if load:
        with chdir(run_directory):
            res = kima.load_results(model)
        return data, model, res

    return data, model


import os, sys
import numpy as np

from .setup_logger import setup_logger
from .utils import ESPRESSO_ADC_issues, ESPRESSO_cryostat_issues


# HARPS started operations on October 1st, 2003
# https://www.eso.org/sci/facilities/lasilla/instruments/harps/news.html
HARPS_start = 52913

# HARPS fiber upgrade (28 May 2015)
# https://www.eso.org/sci/facilities/lasilla/instruments/harps/news/harps_upgrade_2015.html
HARPS_technical_intervention = 57170

# From Lo Curto et al. (2015), The Messenger, vol. 162, p. 9-15  
# On **19 May 2015** HARPS stopped operations and the instrument was opened.
# Installation and alignment of the fibre link lasted roughly one week. On 29
# May, the vacuum vessel was closed and evacuated for the last time. Finally, a
# formal commissioning of the new fibre took place, finishing on **3 June**,
# when the instrument was handed back to Science Operations.
HARPS_technical_intervention_range = (57161, 57176)


# ESPRESSO started operations on October 1st, 2018
# see Pepe et al. (2021, A&A 645, A96)
ESPRESSO_start = 58392

# ESPRESSO fiber link upgrade (1 July 2019)
ESPRESSO_technical_intervention = 58665


def divide_ESPRESSO(self):
    """ Split ESPRESSO data into separate sub ESP18 and ESP19 subsets """
    logger = setup_logger()
    if self._check_instrument('ESPRESSO', strict=False) is None:
        return
    if 'ESPRESSO18' in self.instruments and 'ESPRESSO19' in self.instruments:
        if self.verbose:
            logger.info('ESPRESSO data seems to be split already, doing nothing')
        return

    from .timeseries import RV

    before = self.time < ESPRESSO_technical_intervention
    after = self.time >= ESPRESSO_technical_intervention
    new_instruments = []


    for inst, mask in zip(['ESPRESSO18', 'ESPRESSO19'], [before, after]):
        if not mask.any():
            continue

        _s = RV.from_arrays(self.star, self.time[mask], self.vrad[mask], self.svrad[mask],
                            inst=inst)
        for q in self._quantities:
            setattr(_s, q, getattr(self, q)[mask])
        setattr(self, inst, _s)
        _s._quantities = self._quantities
        _s.mask = self.mask[mask]
        new_instruments.append(inst)

    delattr(self, 'ESPRESSO')
    self.instruments = new_instruments
    self._build_arrays()

    if self.verbose:
        logger.info(f'divided ESPRESSO into {self.instruments}')
    

def divide_HARPS(self):
    """ Split HARPS data into separate sub HARPS03 and HARPS15 subsets """
    logger = setup_logger()
    if self._check_instrument('HARPS', strict=False) is None:
        return
    if 'HARPS03' in self.instruments and 'HARPS15' in self.instruments:
        if self.verbose:
            logger.info('HARPS data seems to be split already, doing nothing')
        return

    from .timeseries import RV

    new_instruments = []
    before = self.time < HARPS_technical_intervention
    if before.any():
        new_instruments += ['HARPS03']

    after = self.time >= HARPS_technical_intervention
    if after.any():
        new_instruments += ['HARPS15']

    for inst, mask in zip(new_instruments, [before, after]):
        _s = RV.from_arrays(self.star, self.time[mask], self.vrad[mask], self.svrad[mask],
                            inst=inst)
        for q in self._quantities:
            setattr(_s, q, getattr(self, q)[mask])
        setattr(self, inst, _s)
        _s._quantities = self._quantities
        _s.mask = self.mask[mask]

    delattr(self, 'HARPS')
    self.instruments = new_instruments
    self._build_arrays()

    if self.verbose:
        logger.info(f'divided HARPS into {self.instruments}')
    

def check(self, instrument):
    return self._check_instrument(instrument)

# HARPS commissioning
def HARPS_commissioning(self, mask=True, plot=True):
    """ Identify and optionally mask points during HARPS commissioning.

    Args:
        mask (bool, optional):
            Whether to mask out the points.
        plot (bool, optional):
            Whether to plot the masked points.
    """
    logger = setup_logger()
    if check(self, 'HARPS') is None:
        return

    affected = self.time < HARPS_start
    total_affected = affected.sum()

    if self.verbose:
        n = total_affected
        logger.info(f"there {'are'[:n^1]}{'is'[n^1:]} {n} frame{'s'[:n^1]} "
                     "during HARPS commissioning")

    if mask:
        self.mask[affected] = False
        self._propagate_mask_changes()

        if plot:
            self.plot(show_masked=True)

    return affected


# HARPS fiber commissioning
def HARPS_fiber_commissioning(self, mask=True, plot=True):
    """ Identify and optionally mask points affected by fiber commissioning (HARPS).

    Args:
        mask (bool, optional):
            Whether to mask out the points.
        plot (bool, optional):
            Whether to plot the masked points.
    """
    logger = setup_logger()
    if check(self, 'HARPS') is None:
        return

    affected = (self.time >= HARPS_technical_intervention_range[0]) & (self.time <= HARPS_technical_intervention_range[1])
    total_affected = affected.sum()

    if self.verbose:
        n = total_affected
        logger.info(f"there {'are'[:n^1]}{'is'[n^1:]} {n} frame{'s'[:n^1]} "
                     "during the HARPS fiber commissioning period")

    if mask:
        self.mask[affected] = False
        self._propagate_mask_changes()

        if plot:
            self.plot(show_masked=True)

    return affected


# ESPRESSO commissioning
def ESPRESSO_commissioning(self, mask=True, plot=True):
    """ Identify and optionally mask points during ESPRESSO commissioning.

    Args:
        mask (bool, optional):
            Whether to mask out the points.
        plot (bool, optional):
            Whether to plot the masked points.
    """
    logger = setup_logger()
    if check(self, 'ESPRESSO') is None:
        return

    affected = self.time < ESPRESSO_start
    total_affected = affected.sum()

    if self.verbose:
        n = total_affected
        logger.info(f"there {'are'[:n^1]}{'is'[n^1:]} {n} frame{'s'[:n^1]} "
                     "during ESPRESSO commissioning")

    if mask:
        self.mask[affected] = False
        self._propagate_mask_changes()

        if plot:
            self.plot(show_masked=True)

    return affected


# ESPRESSO ADC issues
def ADC_issues(self, mask=True, plot=True, check_headers=False):
    """ Identify and optionally mask points affected by ADC issues (ESPRESSO).

    Args:
        mask (bool, optional):
            Whether to mask out the points.
        plot (bool, optional):
            Whether to plot the masked points.
        check_headers (bool, optional):
            Whether to (double-)check the headers for missing/zero keywords.
    """
    logger = setup_logger()
    instruments = self._check_instrument('ESPRESSO')
    
    if instruments is None:
        if self.verbose:
            logger.error(f"ADC_issues: no data from ESPRESSO")
        return

    affected_file_roots = ESPRESSO_ADC_issues()
    file_roots = [os.path.basename(f).replace('.fits', '') for f in self.raw_file]
    intersect = np.in1d(file_roots, affected_file_roots)

    if check_headers:
        from .headers import get_headers
        H = get_headers(self, check_lesta=False, check_exo2=False, instrument='ESPRE')
        badACD2 = np.array([h['*ADC2 RA'][0] for h in H]) == 0
        badACD2 |= np.array([h['*ADC2 SENS1'][0] for h in H]) == 0
        badACD2 |= np.array([h['*ADC2 TEMP'][0] for h in H]) == 0
        intersect = np.logical_or(intersect, badACD2)

    total_affected = intersect.sum()

    if self.verbose:
        n = total_affected
        logger.info(f"there {'are'[:n^1]}{'is'[n^1:]} {n} frame{'s'[:n^1]} "
                     "affected by ADC issues")

    if mask:
        self.mask[intersect] = False
        self._propagate_mask_changes()

        if plot:
            self.plot(show_masked=True)

    return intersect

# ESPRESSO cryostat issues
def blue_cryostat_issues(self, mask=True, plot=True):
    """ Identify and mask points affected by blue cryostat issues (ESPRESSO).

    Args:
        mask (bool, optional): Whether to mask out the points.
        plot (bool, optional): Whether to plot the masked points.
    """
    logger = setup_logger()
    instruments = self._check_instrument('ESPRESSO')
    
    if instruments is None:
        if self.verbose:
            logger.error(f"blue_cryostat_issues: no data from ESPRESSO")
        return

    affected_file_roots = ESPRESSO_cryostat_issues()
    file_roots = [os.path.basename(f).replace('.fits', '') for f in self.raw_file]
    intersect = np.in1d(file_roots, affected_file_roots)

    total_affected = intersect.sum()

    if self.verbose:
        n = total_affected
        logger.info(f"there {'are'[:n^1]}{'is'[n^1:]} {n} frame{'s'[:n^1]} "
                     "affected by blue cryostat issues")

    if mask:
        self.mask[intersect] = False
        self._propagate_mask_changes()

        if plot:
            self.plot(show_masked=True)

    return intersect


def qc_scired_issues(self, plot=False, **kwargs):
    """ Identify and mask points with failed SCIRED QC

    Args:
        plot (bool, optional): Whether to plot the masked points.
    """
    logger = setup_logger()
    from .headers import get_headers

    instruments = self._check_instrument('ESPRESSO')
    
    if instruments is None:
        if self.verbose:
            logger.error(f"no data from ESPRESSO")
            logger.info(f'available: {self.instruments}')
        return

    H = kwargs.get('H', None)
    if H is None:
        H = get_headers(self, check_lesta=False, check_exo2=False, instrument='ESPRE')
    if len(H) == 0:
        if self.verbose:
            logger.warning('this function requires access to headers, but none found')
            logger.warning('trying to download')
        self.download_ccf()
        H = get_headers(self, check_lesta=False, check_exo2=False, instrument='ESPRE')
    
    scired_check = np.array([h['HIERARCH ESO QC SCIRED CHECK'] for h in H])
    affected = scired_check == 0
    n = affected.sum()

    if self.verbose:
        logger.info(f"there {'are'[:n^1]}{'is'[n^1:]} {n} frame{'s'[:n^1]} "
                     "where QC SCIRED CHECK is 0")

    if n == 0:
        return

    self.mask[affected] = False
    self._propagate_mask_changes()

    if plot:
        self.plot(show_masked=True)

    return affected


class ISSUES:
    def known_issues(self, mask=True, plot=False, **kwargs):
        """ Identify and optionally mask known instrumental issues.

        Args:
            mask (bool, optional): Whether to mask out the points.
            plot (bool, optional): Whether to plot the masked points.
        """
        logger = setup_logger()

        functions = (
            ESPRESSO_commissioning,
            ADC_issues,
            blue_cryostat_issues,
            HARPS_commissioning,
            HARPS_fiber_commissioning
        )
        results = []

        for fun in functions:
            try:
                results.append(fun(self, mask, plot, **kwargs))
            except IndexError:
                logger.error('are the data binned? cannot proceed to mask these points...')
        
        results = list(filter(lambda x: x is not None, results))

        try:
            return np.logical_or.reduce(results)
        except UnboundLocalError:
            return



import os
import sys
import tarfile
import collections
from functools import lru_cache, partial
from itertools import islice
import numpy as np

from .setup_logger import setup_logger
from .utils import create_directory, all_logging_disabled, stdout_disabled, timer, tqdm


def load_spectroscopy(user=None):
    logger = setup_logger()
    with all_logging_disabled():
        from dace_query.spectroscopy import SpectroscopyClass, Spectroscopy as default_Spectroscopy
        from dace_query import DaceClass

    from .config import config
    # requesting as public
    if config.request_as_public:
        logger.warning('requesting DACE data as public')
        with all_logging_disabled():
            dace = DaceClass(dace_rc_config_path='none')
        return SpectroscopyClass(dace_instance=dace)
    # DACERC environment variable is set, should point to a dacerc file with credentials
    if 'DACERC' in os.environ:
        dace = DaceClass(dace_rc_config_path=os.environ['DACERC'])
        return SpectroscopyClass(dace_instance=dace)
    # user provided, should be a section in ~/.dacerc
    if user is not None:
        import configparser
        import tempfile
        config = configparser.ConfigParser()
        config.read(os.path.expanduser('~/.dacerc'))
        if user not in config.sections():
            raise ValueError(f'Section for user "{user}" not found in ~/.dacerc')
        with tempfile.NamedTemporaryFile(mode='w', delete=False) as f:
            new_config = configparser.ConfigParser()
            new_config['user'] = config[user]
            new_config.write(f)
        dace = DaceClass(dace_rc_config_path=f.name)
        logger.info(f'using credentials for user {user} in ~/.dacerc')
        return SpectroscopyClass(dace_instance=dace)
    # default
    if not os.path.exists(os.path.expanduser('~/.dacerc')):
        logger.warning('requesting DACE data as public (no .dacerc file found)')
    return default_Spectroscopy


@lru_cache(maxsize=1024)
def get_dace_id(star, verbose=True, raise_error=False):
    logger = setup_logger()
    filters = {"obj_id_catname": {"equal": [star]}}
    try:
        with all_logging_disabled():
            r = load_spectroscopy().query_database(filters=filters, limit=1)
        return str(r['obj_id_daceid'][0])
    except KeyError:
        if verbose:
            logger.error(f"Could not find DACE ID for {star}")
        if not raise_error:
            return None
        raise ValueError from None


def get_arrays(result, latest_pipeline=True, ESPRESSO_mode='HR11', NIRPS_mode='HE', verbose=True):
    logger = setup_logger()
    arrays = []
    instruments = [str(i) for i in result.keys()]

    for inst in instruments:
        pipelines = [str(p) for p in result[inst].keys()]

        # select ESPRESSO mode, which is defined at the level of the pipeline
        if 'ESPRESSO' in inst:
            find_mode = [ESPRESSO_mode in pipe for pipe in pipelines]
            # the mode was not found
            if not any(find_mode):
                if len(pipelines) > 1 and verbose:
                    logger.warning(f'no observations for requested ESPRESSO mode ({ESPRESSO_mode})')
            # the mode was found but do nothing if it's the only one
            elif any(find_mode) and not all(find_mode):
                if verbose:
                    logger.info(f'selecting mode {ESPRESSO_mode} for ESPRESSO')
                i = [i for i, pipe in enumerate(pipelines) if ESPRESSO_mode in pipe][0]
                pipelines = [pipelines[i]]

        # select NIRPS mode
        if 'NIRPS' in inst:
            if any(this_mode := [p for p in pipelines if NIRPS_mode in p]):
                pipelines = this_mode

        if latest_pipeline:
            npipe = len(pipelines)
            if 'NIRPS' in inst and any(['LBL' in p for p in pipelines]):
                # TODO: correctly load both CCF and LBL
                pipelines = [pipelines[1]]
            if 'HARPS' in inst and npipe > 1 and pipelines[1] == pipelines[0] + '-EGGS':
                pipelines = pipelines[:2]
            else:
                pipelines = [pipelines[0]]

            if verbose and npipe > 1:
                logger.info(f'selecting latest pipeline ({pipelines[0]}) for {inst}')

        for pipe in pipelines:
            modes = [m for m in result[inst][pipe].keys()]

            # select NIRPS mode, which is defined at the level of the mode
            if 'NIRPS' in inst and len(modes) > 1:
                if NIRPS_mode in modes:
                    if verbose:
                        logger.info(f'selecting mode {NIRPS_mode} for NIRPS - {pipe}')
                    i = modes.index(NIRPS_mode)
                    modes = [modes[i]]
                else:
                    if verbose:
                        logger.warning(f'no observations for requested NIRPS mode ({NIRPS_mode})')

            # HARPS observations should not be separated by 'mode' if some are
            # done together with NIRPS, but should be separated by 'EGGS' mode
            if 'HARPS' in inst:
                m0 = modes[0]
                data = {
                    k: np.concatenate([result[inst][pipe][m][k] for m in modes])
                    for k in result[inst][pipe][m0].keys()
                }
                if 'HARPS+NIRPS' in modes:
                    arrays.append( ((str(inst), str(pipe), str(m0)), data) )
                    continue

                if 'EGGS+NIRPS' in modes or 'EGGS' in modes:
                    arrays.append( ((str(inst + '_EGGS'), str(pipe), str(m0)), data) )
                    continue

            for mode in modes:
                if 'rjd' not in result[inst][pipe][mode]:
                    logger.error(f"No 'rjd' key for {inst} - {pipe}")
                    raise ValueError

                arrays.append(
                    ((str(inst), str(pipe), str(mode)), result[inst][pipe][mode])
                )

    return arrays

def get_observations_from_instrument(star, instrument, user=None, main_id=None, verbose=True):
    """ Query DACE for all observations of a given star and instrument

    Args:
        star (str):
            name of the star
        instrument (str):
            instrument name
        user (str, optional):
            DACERC user name. Defaults to None.
        main_id (str, optional):
            Simbad main id of target to query DACE id. Defaults to None.
        verbose (bool, optional):
            whether to print warnings. Defaults to True.

    Raises:
        ValueError:
            If query for DACE id fails

    Returns:
        dict:
            dictionary with data from DACE
    """
    Spectroscopy = load_spectroscopy(user)
    found_dace_id = False
    with timer('dace_id query'):
        try:
            dace_id = get_dace_id(star, verbose=verbose, raise_error=True)
            found_dace_id = True
        except ValueError as e:
            if main_id is not None:
                try:
                    dace_id = get_dace_id(main_id, verbose=verbose, raise_error=True)
                    found_dace_id = True
                except ValueError:
                    pass

    if not found_dace_id:
        try:
            with all_logging_disabled():
                result = Spectroscopy.get_timeseries(target=star,
                                                     sorted_by_instrument=True,
                                                     output_format='numpy')
                return result
        except TypeError:
            msg = f'no {instrument} observations for {star}'
            raise ValueError(msg) from None
    if (isinstance(instrument, str)):
        filters = {
            "ins_name": {"contains": [instrument]},
            "obj_id_daceid": {"contains": [dace_id]}
        }
    elif (isinstance(instrument, list)):
        filters = {
            "ins_name": {"contains": instrument},
            "obj_id_daceid": {"contains": [dace_id]}
        }
    with all_logging_disabled():
        result = Spectroscopy.query_database(filters=filters)
    
    if len(result) == 0:
        raise ValueError

    r = {}

    for inst in np.unique(result['ins_name']):
        mask1 = result['ins_name'] == inst
        r[str(inst)] = {}

        key2 = 'ins_drs_version'
        n_key2 = len(np.unique(result[key2][mask1]))
        if len(np.unique(result['pub_bibcode'][mask1])) >= n_key2:
            key2 = 'pub_bibcode'

        for pipe in np.unique(result[key2][mask1]):
            mask2 = mask1 & (result[key2] == pipe)
            r[str(inst)][str(pipe)] = {}

            for ins_mode in np.unique(result['ins_mode'][mask2]):
                mask3 = mask2 & (result['ins_mode'] == ins_mode)
                _nan = np.full(mask3.sum(), np.nan)

                r[str(inst)][str(pipe)][str(ins_mode)] = {
                    'texp': result['texp'][mask3],
                    'bispan': result['spectro_ccf_bispan'][mask3],
                    'bispan_err': result['spectro_ccf_bispan_err'][mask3],
                    'drift_noise': result['spectro_cal_drift_noise'][mask3],
                    'rjd': result['obj_date_bjd'][mask3],
                    'cal_therror': _nan,
                    'fwhm': result['spectro_ccf_fwhm'][mask3],
                    'fwhm_err': result['spectro_ccf_fwhm_err'][mask3],
                    'rv': result['spectro_ccf_rv'][mask3],
                    'rv_err': result['spectro_ccf_rv_err'][mask3],
                    'berv': result['spectro_cal_berv'][mask3],
                    'ccf_noise': _nan,
                    'rhk': result['spectro_analysis_rhk'][mask3],
                    'rhk_err': result['spectro_analysis_rhk_err'][mask3],
                    'contrast': result['spectro_ccf_contrast'][mask3],
                    'contrast_err': result['spectro_ccf_contrast_err'][mask3],
                    'cal_thfile': result['spectro_cal_thfile'][mask3],
                    'spectroFluxSn50': result['spectro_flux_sn50'][mask3],
                    'protm08': result['spectro_analysis_protm08'][mask3],
                    'protm08_err': result['spectro_analysis_protm08_err'][mask3],
                    'caindex': result['spectro_analysis_ca'][mask3],
                    'caindex_err': result['spectro_analysis_ca_err'][mask3],
                    'pub_reference': result['pub_ref'][mask3],
                    'drs_qc': result['spectro_drs_qc'][mask3],
                    'haindex': result['spectro_analysis_halpha'][mask3],
                    'haindex_err': result['spectro_analysis_halpha_err'][mask3],
                    'protn84': result['spectro_analysis_protn84'][mask3],
                    'protn84_err': result['spectro_analysis_protn84_err'][mask3],
                    'naindex': result['spectro_analysis_na'][mask3],
                    'naindex_err': result['spectro_analysis_na_err'][mask3],
                    'snca2': _nan,
                    'mask': result['spectro_ccf_mask'][mask3],
                    'public': result['public'][mask3],
                    'spectroFluxSn20': result['spectro_flux_sn20'][mask3],
                    'sindex': result['spectro_analysis_smw'][mask3],
                    'sindex_err': result['spectro_analysis_smw_err'][mask3],
                    'drift_used': _nan,
                    'ccf_asym': result['spectro_ccf_asym'][mask3],
                    'ccf_asym_err': result['spectro_ccf_asym_err'][mask3],
                    'date_night': result['date_night'][mask3],
                    'raw_file': result['file_rootpath'][mask3],
                    'prog_id': result['prog_id'][mask3],
                    'th_ar': result['th_ar'][mask3],
                    'th_ar1': result['th_ar1'][mask3],
                    'th_ar2': result['th_ar2'][mask3],
                }
    
    # print(r.keys())    
    # print([r[k].keys() for k in r.keys()])
    # print([r[k1][k2].keys() for k1 in r.keys() for k2 in r[k1].keys()])
    return r

def get_observations(star, instrument=None, user=None, main_id=None, verbose=True):
    logger = setup_logger()
    if instrument is None:
        Spectroscopy = load_spectroscopy(user)

        try:
            with stdout_disabled(), all_logging_disabled():
                result = Spectroscopy.get_timeseries(target=star,
                                                     sorted_by_instrument=True,
                                                     output_format='numpy')
        except TypeError:
            if instrument is None:
                msg = f'no observations for {star}'
            else:
                msg = f'no {instrument} observations for {star}'
            raise ValueError(msg) from None
    else:
        try:
            result = get_observations_from_instrument(star, instrument, user, main_id, verbose)
        except ValueError:
            msg = f'no {instrument} observations for {star}'
            raise ValueError(msg) from None

    # defaultdict --> dict
    if isinstance(result, collections.defaultdict):
        result = dict(result)
    for inst in result.keys():
        for pipe in result[inst].keys():
            for mode in result[inst][pipe].keys():
                if isinstance(result[inst][pipe][mode], collections.defaultdict):
                    result[inst][pipe][mode] = dict(result[inst][pipe][mode])
            if isinstance(result[inst][pipe], collections.defaultdict):
                    result[inst][pipe] = dict(result[inst][pipe])
        if isinstance(result[inst], collections.defaultdict):
            result[inst] = dict(result[inst])
    #

    instruments = list(map(str, result.keys()))

    if instrument is not None:
        # select only the provided instrument (if it's there)
        if (isinstance(instrument, str)):
            instruments = [inst for inst in instruments if instrument in inst]
        elif (isinstance(instrument, list)):
            instruments = [inst for inst in instruments if any(i in inst for i in instrument)]
    if len(instruments) == 0:
        if instrument is None:
            msg = f'no observations for {star}'
        else:
            msg = f'no {instrument} observations for {star}'
        raise ValueError(msg)

    # # sort pipelines, being extra careful with HARPS pipeline names
    # # (i.e. ensure that 3.x.x > 3.5)
    # from re import match
    # def cmp(a, b):
    #     if a[0] in ('3.5', '3.5 EGGS') or 'EGGS' in a[0] and match(r'3.\d.\d', b[0]):
    #         return -1
    #     if b[0] in ('3.5', '3.5 EGGS') or 'EGGS' in b[0] and match(r'3.\d.\d', a[0]):
    #         return 1

    #     if a[0] == b[0]:
    #         return 0
    #     elif a[0] > b[0]:
    #         return 1
    #     else:
    #         return -1

    # # sort pipelines, must be extra careful with HARPS/HARPN pipeline version numbers
    # # got here with the help of DeepSeek
    # # from functools import cmp_to_key
    # from re import match
    # def custom_sort_key(s):
    #     s = s[0]
    #     # Check for version number pattern (e.g., 3.2.5 or 3.2.5-EGGS)
    #     version_match = match(r'^(\d+(?:\.\d+)*)(?:[-\s](.*))?$', s)
    #     if version_match:
    #         version_parts = list(map(int, version_match.group(1).split('.')))
    #         if len(version_parts) == 2:
    #             version_parts.insert(1, -1)
    #         # if version_match.group(2) and 'LBL' in version_match.group(2):
    #         #     version_parts.append(-1)
    #         # else:
    #         #     version_parts.append(0)
    #         if version_match.group(2) is None:
    #             version_parts.append('')
    #         else:
    #             version_parts.append(version_match.group(2))
    #         return (0, 1, version_parts)
    #     # Check for scientific reference pattern (e.g., 2004A&A...)
    #     year_match = match(r'^(\d{4})', s)
    #     if year_match:
    #         year = int(year_match.group(1))
    #         return (1, year)
    #     # For all other strings, sort alphabetically
    #     return (2, s)

    def custom_key(val, strip_EGGS=False):
        if strip_EGGS:
            val = val.replace('-EGGS', '').replace(' EGGS', '')
        key = 0
        key -= 1 if '3.5' in val else 0
        key -= 1 if 'EGGS' in val else 0
        key -= 1 if ('UHR' in val or 'MR' in val) else 0
        key -= 1 if 'LBL' in val else 0
        return str(key) if key != 0 else val

    new_result = {}
    for inst in instruments:
        # new_result[inst] = dict(
        #     sorted(result[inst].items(), key=custom_sort_key, reverse=True)
        # )
        if all(['EGGS' in k for k in result[inst].keys()]):
            custom_key = partial(custom_key, strip_EGGS=True)
        # WARNING: not the same as reverse=True (not sure why)
        sorted_keys = sorted(result[inst].keys(), key=custom_key)[::-1]
        new_result[inst] = {}
        for key in sorted_keys:
            new_result[inst][key] = result[inst][key]

    if verbose:
        logger.info('RVs available from')
        with logger.contextualize(indent='  '):
            _inst = ''
            for inst in instruments:
                pipelines = list(new_result[inst].keys())
                max_len = max([len(pipe) for pipe in pipelines])
                for pipe in pipelines:
                    last_pipe = pipe == pipelines[-1]
                    modes = list(new_result[inst][pipe].keys())
                    for mode in modes:
                        N = len(new_result[inst][pipe][mode]['rjd'])
                        # LOG
                        if inst == _inst and last_pipe:
                            logger.info(f'{" ":>12s} └ {pipe:{max_len}s} - {mode} ({N} observations)')
                        elif inst == _inst:
                            logger.info(f'{" ":>12s} ├ {pipe:{max_len}s} - {mode} ({N} observations)')
                        else:
                            logger.info(f'{inst:>12s} ├ {pipe:{max_len}s} - {mode} ({N} observations)')
                            _inst = inst

    return new_result


def check_existing(output_directory, files, type):
    """ Check how many of `files` exist in `output_directory` """
    existing = [
        f.partition('.fits')[0] for f in os.listdir(output_directory)
        if type in f
    ]

    if type == 'S2D':
        existing += [
            f.partition('.fits')[0] for f in os.listdir(output_directory)
            if 'e2ds' in f
        ]   

    # also check for lowercase type
    existing += [
        f.partition('.fits')[0] for f in os.listdir(output_directory)
        if type.lower() in f
    ]

    if os.name == 'nt':  # on Windows, be careful with ':' in filename
        import re
        existing = [re.sub(r'T(\d+)_(\d+)_(\d+)', r'T\1:\2:\3', f) for f in existing]
    
    # remove type of file (e.g. _CCF_A)
    existing = [f.partition('_')[0] for f in existing]
    existing = np.unique(existing)

    missing = []
    for file in files:
        if any(other in file for other in existing):
            continue
        missing.append(file)

    return np.array(missing)

def download(files, type, output_directory, output_filename=None, user=None, quiet=True, pbar=None):
    """ Download files from DACE """
    Spectroscopy = load_spectroscopy(user)
    if isinstance(files, str):
        files = [files]
    if quiet:
        with all_logging_disabled():
            Spectroscopy.download_files(files, file_type=type.lower(),
                                        output_directory=output_directory,
                                        output_filename=output_filename)
    else:
        Spectroscopy.download_files(files, file_type=type.lower(),
                                    output_directory=output_directory,
                                    output_filename=output_filename)
    if pbar is not None:
        pbar.update()


def extract_fits(output_directory, filename=None):
    """ Extract fits files from tar.gz file """
    if filename is None:
        filename = 'spectroscopy_download.tar.gz'
    file = os.path.join(output_directory, filename)
    with tarfile.open(file, "r") as tar:
        files = []
        for member in tar.getmembers():
            if member.isreg():  # skip if the TarInfo is not a file
                member.name = os.path.basename(member.name)  # remove the path
                if os.name == 'nt':  # on Windows, be careful with ':' in filename
                    member.name = member.name.replace(':', '_')
                tar.extract(member, output_directory)
                files.append(member.name)
    os.remove(file)
    return files


def do_symlink_filetype(type, raw_files, output_directory, clobber=False, top_level=None, verbose=True):
    logger = setup_logger()
    terminations = {
        'CCF': '_CCF_A.fits',
        'S1D': '_S1D_A.fits',
        'S2D': '_S2D_A.fits',
    }

    create_directory(output_directory)

    raw_files = np.atleast_1d(raw_files)

    # check existing files
    if not clobber:
        raw_files = check_existing(output_directory, raw_files, type)

    n = raw_files.size

    # any file left?
    if n == 0:
        if verbose:
            logger.info('no files to symlink')
        return

    if verbose:
        msg = f"symlinking {n} {type}s into '{output_directory}'..."
        logger.info(msg)

    for file in tqdm(raw_files):
        if top_level is not None:
            top = file.split('/')[0] + '/'
            if not top_level.endswith('/'):
                top_level = top_level + '/'
            file = file.replace(top, top_level)

        file = file.replace('.fits', terminations[type])

        if os.path.exists(file):
            os.symlink(file, os.path.join(output_directory, os.path.basename(file)))
            # print(file, os.path.join(output_directory, os.path.basename(file)))
        else:
            logger.warning(f'file not found: {file}')


def do_download_filetype(type, raw_files, output_directory, clobber=False, user=None,
                         verbose=True, chunk_size=20, parallel_limit=30):
    """ Download CCFs / S1Ds / S2Ds from DACE """
    logger = setup_logger()
    raw_files = np.atleast_1d(raw_files)

    create_directory(output_directory)

    # check existing files to avoid re-downloading
    if not clobber:
        raw_files = check_existing(output_directory, raw_files, type)
    
    n = raw_files.size

    # any file left to download?
    if n == 0:
        if verbose:
            logger.info('no files to download')
        return

    # avoid an empty chunk
    if chunk_size > n:
        chunk_size = n  

    if verbose:
        if chunk_size < n:
            msg = f"downloading {n} {type}s "
            msg += f"(in chunks of {chunk_size}) "
            msg += f"into '{output_directory}'..."
            logger.info(msg)
        else:
            msg = f"downloading {n} {type}s into '{output_directory}'..."
            logger.info(msg)

    if n < parallel_limit:
        iterator = [raw_files[i:i + chunk_size] for i in range(0, n, chunk_size)]
        for files in tqdm(iterator, total=len(iterator)):
            download(files, type, output_directory, quiet=False, user=user)
            extract_fits(output_directory)

    else:
        def chunker(it, size):
            iterator = iter(it)
            while chunk := list(islice(iterator, size)):
                yield chunk

        chunks = list(chunker(raw_files, chunk_size))
        pbar = tqdm(total=len(chunks))
        it1 = [
            (files, type, output_directory, f'spectroscopy_download{i+1}.tar.gz', user, True, pbar)
            for i, files in enumerate(chunks)
        ]
        it2 = [(output_directory, f'spectroscopy_download{i+1}.tar.gz') for i in range(len(chunks))]

        # import multiprocessing as mp
        # with mp.Pool(4) as pool:
        from multiprocessing.pool import ThreadPool

        with ThreadPool(4) as pool:
            pool.starmap(download, it1)
            pool.starmap(extract_fits, it2)
            print('')

    sys.stdout.flush()
    logger.info('extracted .fits files')


# def do_download_s1d(raw_files, output_directory, clobber=False, verbose=True):
#     """ Download S1Ds from DACE """
#     raw_files = np.atleast_1d(raw_files)

#     create_directory(output_directory)

#     # check existing files to avoid re-downloading
#     if not clobber:
#         raw_files = check_existing(output_directory, raw_files, 'S1D')

#     # any file left to download?
#     if raw_files.size == 0:
#         if verbose:
#             logger.info('no files to download')
#         return

#     if verbose:
#         n = raw_files.size
#         logger.info(f"Downloading {n} S1Ds into '{output_directory}'...")

#     download(raw_files, 's1d', output_directory)

#     if verbose:
#         logger.info('Extracting .fits files')

#     extract_fits(output_directory)


# def do_download_s2d(raw_files, output_directory, clobber=False, verbose=True):
#     """ Download S2Ds from DACE """
#     raw_files = np.atleast_1d(raw_files)

#     create_directory(output_directory)

#     # check existing files to avoid re-downloading
#     if not clobber:
#         raw_files = check_existing(output_directory, raw_files, 'S2D')

#     # any file left to download?
#     if raw_files.size == 0:
#         if verbose:
#             logger.info('no files to download')
#         return

#     if verbose:
#         n = raw_files.size
#         logger.info(f"Downloading {n} S2Ds into '{output_directory}'...")

#     download(raw_files, 's2d', output_directory)

#     if verbose:
#         logger.info('Extracting .fits files')

#     extracted_files = extract_fits(output_directory)
#     return extracted_files



import numpy as np

def wmean(a, e):
    """Weighted mean of array `a`, with uncertainties given by `e`.
    The weighted mean is calculated using 1/e**2 as weights.

    Args:
        a (array): Array containing data
        e (array): Uncertainties on `a`
    """
    if (e == 0).any():
        raise ZeroDivisionError
    if (e < 0).any():
        raise ValueError
    if (a.shape != e.shape):
        raise ValueError
    return np.average(a, weights=1 / e**2)

def rms(a, ignore_nans=False):
    """ Root mean square of array `a`

    Args:
        a (array): Array containing data
    """
    if ignore_nans:
        a = a[~np.isnan(a)]
    if len(a) == 0:
        return np.nan
    return np.sqrt((a**2).mean())

def wrms(a, e, ignore_nans=False):
    """ Weighted root mean square of array `a`, with uncertanty given by `e`.
    The weighted rms is calculated using the weighted mean, where the weights
    are equal to 1/e**2.
    
    Args:
        a (array): Array containing data 
        e (array): Uncertainties on `a`
    """
    if ignore_nans:
        nans = np.logical_or(np.isnan(a), np.isnan(e))
        a = a[~nans]
        e = e[~nans]
    if (e == 0).any():
        raise ZeroDivisionError('uncertainty cannot be zero')
    if (e < 0).any():
        raise ValueError('uncertainty cannot be negative')
    if (a.shape != e.shape):
        raise ValueError('arrays must have the same shape')
    w = 1 / e**2
    return np.sqrt(np.sum(w * (a - np.average(a, weights=w))**2) / sum(w))


def sigmaclip_median(a, low=4.0, high=4.0):
    """
    Same as scipy.stats.sigmaclip but using the median and median absolute
    deviation instead of the mean and standard deviation.

    Args:
        a (array): Array containing data
        low (float): Number of MAD to use for the lower clipping limit
        high (float): Number of MAD to use for the upper clipping limit
    Returns:
        SigmaclipResult: Object with the following attributes:
            - `clipped`: Masked array of data
            - `lower`: Lower clipping limit
            - `upper`: Upper clipping limit
    """
    from scipy.stats import median_abs_deviation
    from scipy.stats._stats_py import SigmaclipResult
    c = np.asarray(a).ravel()
    delta = 1
    while delta:
        c_mad = median_abs_deviation(c)
        c_median = np.median(c)
        size = c.size
        critlower = c_median - c_mad * low
        critupper = c_median + c_mad * high
        c = c[(c >= critlower) & (c <= critupper)]
        delta = size - c.size

    return SigmaclipResult(c, critlower, critupper)


from multiprocessing.pool import ThreadPool
import re
import requests
from io import StringIO

import numpy as np

from .setup_logger import setup_logger

URL_CCF = "http://atlas.obs-hp.fr/sophie/sophie.cgi?n=sophiecc&ob=date&a=t&o={target}"
URL_HEADER = "http://atlas.obs-hp.fr/sophie/sophie.cgi?n=sophiecc&c=i&z=fd&a=t&o=sophie:[ccf,{seq},{mask},0]"

def extract_keyword(keyword, text, raise_error=True):
    for line in text.splitlines():
        if keyword in line:
            value = re.findall(fr'{keyword}\s+([\'\w\d.]+)', line)[0]
            value = value.replace("'", "")
            try:
                return float(value)
            except ValueError:
                return value
    if raise_error:
        raise KeyError(f'Keyword {keyword} not found')

def query_sophie_archive(star: str, verbose=True):
    from .timeseries import RV
    logger = setup_logger()

    resp = requests.get(URL_CCF.format(target=star))
    if 'leda did not return a position for the name' in resp.text:
        raise ValueError(f'no SOPHIE observations for {star}')

    data = np.genfromtxt(StringIO(resp.text), dtype=None, usecols=(0, 4),
                         names=("seq", "mask"))
    
    if verbose:
        logger.info(f'found {len(data)} SOPHIE observations for {star}')

    urls = [URL_HEADER.format(seq=seq, mask=mask) for seq, mask in data]
    with ThreadPool(8) as pool:
        responses = pool.map(requests.get, urls)

    bjd, vrad, svrad = [], [], []
    fwhm, contrast = [], []
    ccf_mask = []
    _quantities = []
    errors = []

    for i, resp in enumerate(responses):
        if resp.text == '':
            errors.append(i)
            continue
        
        try:
            t, v = map(lambda k: extract_keyword(k, resp.text), 
                       ("OHP DRS BJD", "OHP DRS CCF RV"))
        except KeyError:
            errors.append(i)
            continue
        else:
            bjd.append(t)
            vrad.append(v)
        
        try:
            svrad.append(extract_keyword("OHP DRS CCF ERR", resp.text))
        except KeyError:
            try:
                svrad.append(1e-3 * extract_keyword("OHP DRS DVRMS", resp.text))
            except KeyError:
                bjd.pop(-1)
                vrad.pop(-1)
                errors.append(i)
                continue

        fwhm.append(extract_keyword('OHP DRS CCF FWHM', resp.text))
        _quantities.append('fwhm')

        contrast.append(extract_keyword('OHP DRS CCF CONTRAST', resp.text))
        _quantities.append('contrast')

        ccf_mask.append(extract_keyword('OHP DRS CCF MASK', resp.text))
        _quantities.append('ccf_mask')

    if len(errors) > 0:
        logger.warning(f'Could not retrieve {len(errors)} observation'
                       f'{"s" if len(errors) > 1 else ""}')

    bjd = np.array(bjd) - 2400000.5

    s = RV.from_arrays(star, bjd, vrad, svrad, 'SOPHIE',
                       fwhm=fwhm, fwhm_err=2*np.array(svrad),
                       contrast=contrast, 
                       ccf_mask=ccf_mask)
    s.units = 'km/s'

    # strings
    for q in ['date_night', 'prog_id', 'raw_file', 'pub_reference']:
        setattr(s, q, np.full(bjd.size, ''))
        _quantities.append(q)

    s._quantities = np.array(_quantities)

    setattr(s, 'SOPHIE', s)
    s._child = False
    s.verbose = False
    s._build_arrays()
    s.change_units('m/s')
    s.verbose = verbose

    return s
    



import pytest
import os

def test_import():
    import arvi
    from arvi import RV


def test_import_object():
    from arvi import _51Peg
    from arvi import HD10180

def test_import_object_public():
    from arvi import config

    config.request_as_public = True
    
    from arvi import _51Peg
    from arvi import HD10180

    config.request_as_public = False


def test_fancy_import_off():
    from arvi import config

    from arvi import HD10180

    config.fancy_import = False

    with pytest.raises(ImportError):
        from arvi import HD69830




from tqdm import tqdm
from astropy.io import fits
import iCCF

from . import config

def get_headers(self, check_lesta=False, lesta_username=config.username,
                check_exo2=False, instrument=None):
    try:
        import paramiko
    except ImportError:
        raise ImportError("paramiko is not installed. Please install it with 'pip install paramiko'")

    H = []

    if check_lesta:
        with paramiko.SSHClient() as ssh:
            ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            ssh.connect("lesta02.astro.unige.ch", username=lesta_username)
            sftp = ssh.open_sftp()

            pbar = tqdm(self.raw_file, total=len(self.raw_file), unit='file', desc='Reading headers')
            for f in pbar:
                f = f.replace('espresso/', '/projects/astro/ESPRESSODRS/')
                f = f.replace('nirps/', '/projects/astro/NIRPSDRS/')
                f = f.replace('.fits', '_CCF_A.fits')#.replace(':', r'\:')
                with sftp.open(f) as fp:
                    header = fits.getheader(fp)
                    H.append(header)

    if len(H) == 0 and check_exo2:
        raise NotImplementedError('getting headers from exo2 not yet implemented')

    if len(H) == 0:
        self.download_ccf()
        if instrument is None:
            I = iCCF.from_file(f'{self.star}_downloads/*CCF_A.fits')
        else:
            I = iCCF.from_file(f'{self.star}_downloads/r.{instrument}.*CCF_A.fits',
                               guess_instrument='HARPS' not in instrument)
        H = [i.HDU[0].header for i in I]

    # sort by BJD
    H = sorted(H, key=lambda x: x['HIERARCH ESO QC BJD'])

    return H




__all__ = ['RV', 'config', 'simbad', 'gaia']

from importlib.metadata import version, PackageNotFoundError
try:
    __version__ = version("arvi")
except PackageNotFoundError:
    # package is not installed
    pass

from .config import config
from .simbad_wrapper import simbad
from .gaia_wrapper import gaia

from .timeseries import RV

def __getattr__(name: str):
    if not config.fancy_import:
        raise AttributeError
    
    if name in (
        '_ipython_canary_method_should_not_exist_',
        '_ipython_display_',
        '_repr_mimebundle_',
        '__wrapped__'
    ):
        return

    try:
        globals()[name] = RV(name)
        return globals()[name]
    except ValueError as e:
        raise AttributeError(e)




import numpy as np
from astropy.constants import G
from astropy import units

# Coeffcients to be used in the analytical expression to calculate habitable
# zone flux boundaries
seffsun = [1.776, 1.107, 0.356, 0.320, 1.188, 0.99]
a = [2.136e-4, 1.332e-4, 6.171e-5, 5.547e-5, 1.433e-4, 1.209e-4]
b = [2.533e-8, 1.580e-8, 1.698e-9, 1.526e-9, 1.707e-8, 1.404e-8]
c = [-1.332e-11, -8.308e-12, -3.198e-12, -2.874e-12, -8.968e-12, -7.418e-12]
d = [-3.097e-15, -1.931e-15, -5.575e-16, -5.011e-16, -2.084e-15, -1.713e-15]
seffsun = np.array(seffsun)
a, b, c, d = np.array([a, b, c, d])


def getHZ(teff, lum, which='conservative'):
    """ 
    Calculate the limits of the HZ, according to Kopparapu et al.(2013).

    Parameters
    ----------
    teff : float
        Stellar effective temperature
    lum : float
        Stellar luminosity (in units of the solar luminosity)
    which : str
        Either 'conservative' or 'optimistic' habitable zone

    Returns
    -------
    innerHZf, outerHZf : floats
        Inner and outer limits of the HZ [in stellar flux compared to the Sun].
    innerHZd, outerHZd : floats
        Inner and outer limits of the HZ [in AU].
    """
    tstar = teff - 5780
    seff = seffsun + a * tstar + b * tstar**2 + c * tstar**3 + d * tstar**4
    dist = np.sqrt(lum / seff)

    # the order of the coefficients is
    # recentVenus
    # runawayGreenhouse
    # maxGreenhouse
    # earlyMars
    # runaway5Me
    # runaway10Me

    if which in ('conservative', 'con'):
        return (seff[1], seff[2]), (dist[1], dist[2])
    elif which in ('optimistic', 'opt'):
        return (seff[0], seff[3]), (dist[0], dist[3])
    else:
        raise ValueError(f'Could not recognise `which={which}`')


def getHZ_period(teff, Mstar, Mplanet, lum=1, Mplanet_units='earth',
                 which='conservative'):
    """
    Calculate the period limits of the HZ.

    Parameters
    ----------
    teff : float
        Stellar effective temperature
    Mstar : float
        Mass of the star
    Mplanet : float
        Mass of the planet
    lum : float
        Stellar luminosity (in units of the solar luminosity)
    Mplanet_units : str
        Units of the planet mass, 'earth' or 'jupiter'
    which : str
        Either 'conservative' or 'optimistic' habitable zone
    
    Returns
    -------
    innerHZd, outerHZd : floats
        Inner and outer limits of the HZ [in days].
    """
    # this function is nothing more than Kepler's 3rd law

    f = 4 * np.pi**2 / G
    
    HZa = np.array(getHZ(teff, lum, which)[1]) * units.AU

    Mstar = Mstar * units.solMass

    if Mplanet_units.lower() == 'earth':
        Mplanet = Mplanet * units.earthMass
    elif Mplanet_units.lower() in ('jupiter', 'jup'):
        Mplanet = Mplanet * units.jupiterMass

    return np.sqrt(f * HZa**3 / (Mstar + Mplanet)).to(units.day)


from functools import partial
import numpy as np
from astropy.timeseries import LombScargle

from .setup_logger import setup_logger


sine_line = None
residual_gls = None
def sine_picker(event, self, fig, ax, ax1):
    from .timeseries import fit_sine
    global sine_line, residual_gls
    if sine_line is not None:
        sine_line[0].remove()
    if residual_gls is not None:
        residual_gls[0].remove()
    xdata, ydata = event.artist.get_data()
    ind = event.ind
    period = xdata[ind][0]
    p, sine = fit_sine(self.mtime, self.mvrad, self.msvrad, period=period)
    tt = np.linspace(self.mtime.min(), self.mtime.max(), 100)
    tt -= 50000
    sine_line = ax1.plot(tt, sine(tt), 'k')
    #
    f, p = LombScargle(self.mtime, self.mvrad - sine(self.mtime), self.msvrad).autopower()
    residual_gls = ax.semilogx(1/f, p, 'r')
    fig.canvas.draw_idle()


class REPORTS:
    def summary(self, add_ccf_mask=True, add_prog_id=False, **kwargs):
        from .utils import pretty_print_table
        if isinstance(self, list):
            selfs = self
        else:
            selfs = [self]
        
        rows = []
        for self in selfs:
            rows.append([self.star] + [''] * len(self.instruments) + [''])
            rows.append([''] + self.instruments + ['full'])
            rows.append(['N'] + list(self.NN.values()) + [self.N])
            rows.append(['RV span'] + [np.ptp(s.mvrad).round(3) for s in self] + [np.ptp(self.mvrad).round(3)])
            rows.append(['RV std'] + [s.mvrad.std().round(3) for s in self] + [self.mvrad.std().round(3)])
            rows.append(['eRV mean'] + [s.msvrad.mean().round(3) for s in self] + [self.msvrad.mean().round(3)])

            if add_ccf_mask:
                if hasattr(self, 'ccf_mask'):
                    row = ['CCF mask']
                    for inst in self.instruments:
                        row.append(', '.join(np.unique(getattr(self, inst).ccf_mask)))
                    row.append('')
                    rows.append(row)

            if add_prog_id:
                row = ['prog ID']
                for inst in self.instruments:
                    p = ', '.join(np.unique(getattr(self, inst).prog_id))
                    row.append(p)
                rows.append(row)

        return pretty_print_table(rows, **kwargs)


    def summary_one_instrument(self):
        from .utils import pretty_print_table
        if isinstance(self, list):
            selfs = self
        else:
            selfs = [self]

        rows = []
        rows.append(['', 'N', 'Δt', 'RV std', 'median σRV'])
        for self in selfs:
            rows.append([
                self.star, 
                self.N, 
                int(np.ptp(self.mtime).round(0)),
                np.std(self.mvrad).round(1),
                np.median(self.msvrad).round(1),
            ])

        pretty_print_table(rows)


    def summary_stellar(self, show_ra_dec=True, show_pm=True, **kwargs):
        from .utils import pretty_print_table, get_ra_sexagesimal, get_dec_sexagesimal
        if isinstance(self, list):
            selfs = self
        else:
            selfs = [self]
        


        rows = []
        rows.append(['star'] + [self.star for self in selfs])
        if show_ra_dec:
            rows.append(['RA'] + [get_ra_sexagesimal(self.simbad.ra) for self in selfs])
            rows.append(['DEC'] + [get_dec_sexagesimal(self.simbad.dec) for self in selfs])
        
        if show_pm:
            rows.append(['pm ra'] + [(self.gaia or self.simbad).pmra for self in selfs])
            rows.append(['pm dec'] + [(self.gaia or self.simbad).pmdec for self in selfs])

        rows.append(['π (mas)'] + [(self.gaia or self.simbad).plx for self in selfs])
        # 
        rows.append(['B (mag)'] + [self.simbad._B for self in selfs])
        rows.append(['V (mag)'] + [self.simbad._V for self in selfs])
        rows.append(['B-V'] + [self.simbad._B - self.simbad._V for self in selfs])
        rows.append(["log R'HK"] + [np.nanmean(self.rhk).round(2) for self in selfs])
            # rows.append([''] + self.instruments + ['full'])
            # rows.append(['N'] + list(self.NN.values()) + [self.N])
            # rows.append(['RV span'] + [np.ptp(s.mvrad).round(3) for s in self] + [np.ptp(self.mvrad).round(3)])
            # rows.append(['RV std'] + [s.mvrad.std().round(3) for s in self] + [self.mvrad.std().round(3)])
            # rows.append(['eRV mean'] + [s.msvrad.mean().round(3) for s in self] + [self.msvrad.mean().round(3)])

        return pretty_print_table(rows, **kwargs)




    def report(self, save=None):
        import matplotlib.pyplot as plt
        import matplotlib.gridspec as gridspec
        from matplotlib.backends.backend_pdf import PdfPages
        logger = setup_logger()

        # size = A4
        size = 8.27, 11.69
        fig = plt.figure(figsize=size, constrained_layout=True)
        gs = gridspec.GridSpec(5, 3, figure=fig, height_ratios=[2, 2, 1, 1, 0.1])

        # first row, all columns
        ax1 = plt.subplot(gs[0, :])

        title = f'{self.star}'
        ax1.set_title(title, loc='left', fontsize=14)
        # ax1.set_title(r"\href{http://www.google.com}{link}", color='blue',
        #               loc='center')

        if self._did_adjust_means:
            title = '(instrument means subtracted) '
        else:
            title = ''
        title += f'V={self.simbad.V}, {self.simbad.sp_type}'
        ax1.set_title(title, loc='right', fontsize=12)

        self.plot(ax=ax1, N_in_label=True, tooltips=False, remove_50000=True)


        ax1.legend().remove()
        legend_ax = plt.subplot(gs[1, -1])
        legend_ax.axis('off')
        leg = plt.legend(*ax1.get_legend_handles_labels(),
                        prop={'family': 'monospace'})
        legend_ax.add_artist(leg)
        
        second_legend = f'rms  : {self.rms:.2f} {self.units}'
        second_legend += f'\nerror: {self.error:.2f} {self.units}'

        second_legend += f'\n\nCCF masks: {np.unique(self.ccf_mask)}'

        legend_ax.legend([],
                        title=second_legend,
                        loc='lower right', frameon=False,
                        prop={'family': 'monospace'})

        ax2 = plt.subplot(gs[1, :-1])
        self.gls(ax=ax2, picker=True)

        ax3 = plt.subplot(gs[2, :-1])
        self.plot_fwhm(ax=ax3, tooltips=False, remove_50000=True)
        ax3.legend().remove()
        ax3p = plt.subplot(gs[2, -1])
        self.gls_fwhm(ax=ax3p, picker=False)

        ax4 = plt.subplot(gs[3, :-1])
        self.plot_rhk(ax=ax4, tooltips=False, remove_50000=True)
        ax4.legend().remove()
        ax4p = plt.subplot(gs[3, -1])
        self.gls_rhk(ax=ax4p, picker=False)


        if save is None:
            fig.canvas.mpl_connect(
                'pick_event',
                partial(sine_picker, self=self, fig=fig, ax=ax2, ax1=ax1))

        if save is not None:
            if save is True:
                save = f'report_{"".join(self.star.split())}.pdf'

            if save.endswith('.png'):
                fig.savefig(save)
            else:
                with PdfPages(save) as pdf:
                    #pdf.attach_note('hello', positionRect=[5, 15, 20, 30])

                    if self.verbose:
                        logger.info(f'saving to {save}')
                    pdf.savefig(fig)
                # os.system(f'evince {save} &')

        return fig


import numpy as np

from .setup_logger import setup_logger

###############################################################################
# the following is mostly a copy of the scipy implementation of
# binned_statistic and binned_statistic_dd
# but allowing for a weights parameter

## careful here!
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)
xrange = range


def binned_statistic(x, values, statistic='mean', bins=10, range=None,
                     weights=None):
    try:
        N = len(bins)
    except TypeError:
        N = 1

    if N != 1:
        bins = [np.asarray(bins, float)]

    if range is not None:
        if len(range) == 2:
            range = [range]

    medians, edges, binnumbers = binned_statistic_dd(
        [x], values, statistic, bins, range, None, weights)
    return medians, edges[0], binnumbers


def binned_statistic_dd(sample, values, statistic='mean', bins=10, range=None,
                        expand_binnumbers=False, weights=None):
    from numpy.testing import suppress_warnings
    known_stats = [
        'mean', 'median', 'count', 'sum', 'std', 'min', 'max', 'ptp'
    ]
    if not callable(statistic) and statistic not in known_stats:
        raise ValueError('invalid statistic %r' % (statistic, ))

    # `Ndim` is the number of dimensions (e.g. `2` for `binned_statistic_2d`)
    # `Dlen` is the length of elements along each dimension.
    # This code is based on np.histogramdd
    try:
        # `sample` is an ND-array.
        Dlen, Ndim = sample.shape
    except (AttributeError, ValueError):
        # `sample` is a sequence of 1D arrays.
        sample = np.atleast_2d(sample).T
        Dlen, Ndim = sample.shape

    # Store initial shape of `values` to preserve it in the output
    values = np.asarray(values)
    input_shape = list(values.shape)
    # Make sure that `values` is 2D to iterate over rows
    values = np.atleast_2d(values)
    if weights is not None:
        weights = np.atleast_2d(weights)
    Vdim, Vlen = values.shape

    # Make sure `values` match `sample`
    if (statistic != 'count' and Vlen != Dlen):
        raise AttributeError('The number of `values` elements must match the '
                             'length of each `sample` dimension.')

    nbin = np.empty(Ndim, int)  # Number of bins in each dimension
    edges = Ndim * [None]  # Bin edges for each dim (will be 2D array)
    dedges = Ndim * [None]  # Spacing between edges (will be 2D array)

    try:
        M = len(bins)
        if M != Ndim:
            raise AttributeError('The dimension of bins must be equal '
                                 'to the dimension of the sample x.')
    except TypeError:
        bins = Ndim * [bins]

    # Select range for each dimension
    # Used only if number of bins is given.
    if range is None:
        smin = np.atleast_1d(np.array(sample.min(axis=0), float))
        smax = np.atleast_1d(np.array(sample.max(axis=0), float))
    else:
        smin = np.zeros(Ndim)
        smax = np.zeros(Ndim)
        for i in xrange(Ndim):
            smin[i], smax[i] = range[i]

    # Make sure the bins have a finite width.
    for i in xrange(len(smin)):
        if smin[i] == smax[i]:
            smin[i] = smin[i] - .5
            smax[i] = smax[i] + .5

    # Create edge arrays
    for i in xrange(Ndim):
        if np.isscalar(bins[i]):
            nbin[i] = bins[i] + 2  # +2 for outlier bins
            edges[i] = np.linspace(smin[i], smax[i], nbin[i] - 1)
        else:
            edges[i] = np.asarray(bins[i], float)
            nbin[i] = len(edges[i]) + 1  # +1 for outlier bins
        dedges[i] = np.diff(edges[i])

    nbin = np.asarray(nbin)

    # Compute the bin number each sample falls into, in each dimension
    sampBin = [np.digitize(sample[:, i], edges[i]) for i in xrange(Ndim)]

    # Using `digitize`, values that fall on an edge are put in the right bin.
    # For the rightmost bin, we want values equal to the right
    # edge to be counted in the last bin, and not as an outlier.
    for i in xrange(Ndim):
        # Find the rounding precision
        decimal = int(-np.log10(dedges[i].min())) + 6
        # Find which points are on the rightmost edge.
        on_edge = np.where(
            np.around(sample[:, i], decimal) == np.around(
                edges[i][-1], decimal))[0]
        # Shift these points one bin to the left.
        sampBin[i][on_edge] -= 1

    # Compute the sample indices in the flattened statistic matrix.
    binnumbers = np.ravel_multi_index(sampBin, nbin)

    result = np.empty([Vdim, nbin.prod()], float)

    if statistic == 'mean':
        result.fill(np.nan)
        flatcount = np.bincount(binnumbers, None)
        a = flatcount.nonzero()
        for vv in xrange(Vdim):
            flatsum = np.bincount(binnumbers, values[vv])
            result[vv, a] = flatsum[a] / flatcount[a]
    elif statistic == 'std':
        result.fill(0)
        flatcount = np.bincount(binnumbers, None)
        a = flatcount.nonzero()
        for vv in xrange(Vdim):
            flatsum = np.bincount(binnumbers, values[vv])
            flatsum2 = np.bincount(binnumbers, values[vv]**2)
            result[vv, a] = np.sqrt(flatsum2[a] / flatcount[a] -
                                    (flatsum[a] / flatcount[a])**2)
    elif statistic == 'count':
        result.fill(0)
        flatcount = np.bincount(binnumbers, None)
        a = np.arange(len(flatcount))
        result[:, a] = flatcount[np.newaxis, :]
    elif statistic == 'sum':
        result.fill(0)
        for vv in xrange(Vdim):
            flatsum = np.bincount(binnumbers, values[vv])
            a = np.arange(len(flatsum))
            result[vv, a] = flatsum
    elif statistic == 'median':
        result.fill(np.nan)
        for i in np.unique(binnumbers):
            for vv in xrange(Vdim):
                result[vv, i] = np.median(values[vv, binnumbers == i])
    elif statistic == 'min':
        result.fill(np.nan)
        for i in np.unique(binnumbers):
            for vv in xrange(Vdim):
                result[vv, i] = np.min(values[vv, binnumbers == i])
    elif statistic == 'max':
        result.fill(np.nan)
        for i in np.unique(binnumbers):
            for vv in xrange(Vdim):
                result[vv, i] = np.max(values[vv, binnumbers == i])
    elif statistic == 'ptp':
        result.fill(np.nan)
        for i in np.unique(binnumbers):
            for vv in xrange(Vdim):
                result[vv, i] = np.ptp(values[vv, binnumbers == i])
    elif callable(statistic):
        with np.errstate(invalid='ignore'), suppress_warnings() as sup:
            sup.filter(RuntimeWarning)
            try:
                null = statistic([], weights=[])
            except:
                null = np.nan
        result.fill(null)
        for i in np.unique(binnumbers):
            for vv in xrange(Vdim):
                if weights is None:
                    result[vv, i] = statistic(values[vv, binnumbers == i])
                else:
                    result[vv, i] = statistic(values[vv, binnumbers == i],
                                              weights[vv, binnumbers == i])

    # Shape into a proper matrix
    result = result.reshape(np.append(Vdim, nbin))

    # Remove outliers (indices 0 and -1 for each bin-dimension).
    core = tuple([slice(None)] + Ndim * [slice(1, -1)])
    result = result[core]

    # Unravel binnumbers into an ndarray, each row the bins for each dimension
    if (expand_binnumbers and Ndim > 1):
        binnumbers = np.asarray(np.unravel_index(binnumbers, nbin))

    if np.any(result.shape[1:] != nbin - 2):
        raise RuntimeError('Internal Shape Error')

    # Reshape to have output (`reulst`) match input (`values`) shape
    result = result.reshape(input_shape[:-1] + list(nbin - 2))

    return result, edges, binnumbers


# # put back the documentation
# from scipy.stats import binned_statistic as old_binned_statistic,\
#                         binned_statistic_dd as old_binned_statistic_dd
# doc1 = old_binned_statistic.__doc__
# doc2 = old_binned_statistic_dd.__doc__
# binned_statistic.__doc__ = doc1
# binned_statistic_dd.__doc__ = doc2

###############################################################################


def binRV(time, rv, err=None, stat='wmean', tstat='wmean', estat='addquad',
          remove_nans=True, binning_indices=False, binning_bins=False,
          n_consecutive=None, consecutive_step=None, seed=None):
    """ Bin a dataset of radial-velocity observations.

    Parameters
    ----------
    time : array
        The array of times where the radial velocity is measured. This function
        does nightly binning, based on the integer part of this array.
    rv : array
        The radial-velocity values.
    err : array (optional)
        The uncertainty on the radial-velocity values.
    stat : string or callable (optional)
        The statistic to compute for the radial velocities. By default, this is
        'mean' if the `err` array is not provided and weighted mean ('wmean') 
        if `err` is provided. In that case, the routine does inverse variance
        averaging, using 1/`err`**2 as weights. See other available statistics
        on the `binned_statistic` function.
    tstat : string or callable (optional)
        The statistic to compute for the times. The default is the same as for
        the radial velocities. See other available statistics on the
        `binned_statistic` function.
    estat : string or callable (optional)
        The statistic to compute for the errors. The default is to add them
        quadratically and divide by the number. See other available statistics
        on the `binned_statistic` function.
    remove_nans : bool (optional, default True)
        Whether NaN values in the binned times should be removed
    binning_indices : bool (optional, default False)
        If True, don't actually do any binning, just return the indices that
        will nightly bin the `time` array.
    binning_bins : bool (optional, default=False)
        If True, don't actually do any binning, just return the bins that can be
        passed to binned_statistic to nightly bin the `time` array.
    n_consecutive : int (optional)
        If provided, only `n_consecutive` points within each bin will be used
        for the calculation (if more than that number are available). Which
        points are used is chosen randomly.
    consecutive_step : float (optional)
        If `n_consecutive` is provided, `consecutive_step` defines the maximum
        time step between two points that would be considered consecutive.

    Notes
    -----
    Arguably, the most justified way to perform binning is to use the defaults,
    with times, radial velocities and errors as input. This will lead to a
    weighted average of the times, a weighted average of the radial velocities
    and quadratically added errors.
    """

    # "nightly" binning, based on the integer part of the time array
    intt = time.astype(int)
    _, indices = np.unique(intt, return_index=True)

    if binning_indices:
        return indices

    # include the last time in the array of bins, so we don't miss a point
    # but be careful: if it is alone, perturb the last time by a small ammount
    # so that the difference time[-1] - bins[-1] is not zero
    if indices[-1] == time.size - 1:
        bins = np.r_[time[indices], time[-1] + 1e-10]
    else:
        bins = np.r_[time[indices], time[-1]]

    if binning_bins:
        return bins

    if n_consecutive and not consecutive_step:
        print('Warning: ignoring `n_consecutive` because `consecutive_step` '
              'was not provided')

    elif n_consecutive and consecutive_step:
        # easiest solution is to rebuild the time, rv, and err arrays with
        # only the randomly selected n_consecutive points for each bin
        np.random.seed(seed)

        def consecutive(a, step):
            return (np.ediff1d(a) < step).all()

        newtime = []
        newrv = []
        newerr = []
        digits = np.digitize(time, bins)
        random_choices = []
        for d in np.unique(digits):
            m = np.where(digits == d)[0]
            poss = [
                slice(i, i + n_consecutive)
                for i in range(time[m].size - n_consecutive - 1)
                if consecutive(time[m][slice(i, i + n_consecutive)], consecutive_step)
            ]
            if len(poss) > 0:
                sl = np.random.choice(poss)
                newtime.append(time[m][sl])
                newrv.append(rv[m][sl])
                if err is not None:
                    newerr.append(err[m][sl])
                random_choices.append(
                    slice(m[0] + sl.start, m[0] + sl.stop, sl.step))

        time = np.array(newtime).flatten()
        rv = np.array(newrv).flatten()
        if err is not None:
            err = np.array(newerr).flatten()

    # weighted mean of a
    def wmean(a, e):
        return np.ma.average(a, weights=1 / e**2)

    # bin the RVs
    if (err is not None) and (stat == 'wmean'):
        stat = wmean  # default is weighted mean

    if (err is None) and (stat == 'wmean'):
        stat = 'mean'

    brv = binned_statistic(time, rv, statistic=stat, bins=bins, range=None,
                           weights=err)

    # bin the times
    if (err is not None) and (tstat == 'wmean'):
        tstat = wmean  # default is weighted mean

    if (err is None) and (tstat == 'wmean'):
        tstat = 'mean'

    times = binned_statistic(time, time, statistic=tstat, bins=bins,
                             range=None, weights=err)
    # if there are errors, bin them too
    if err is not None:
        if estat == 'addquad':
            # this function adds the elements of `a` quadratically
            add_quadratically = lambda a: np.sqrt(np.add.reduce(a**2))
            # this function then divides by the number of elements
            mean_quadratically = lambda a: add_quadratically(a) / a.size
            # the two previous functions are only separated for clarity
            estat = mean_quadratically
            errors = binned_statistic(time, err, statistic=estat, bins=bins)
        else:
            errors = binned_statistic(time, err, statistic=estat, bins=bins)
    else:
        errors = [None]

    btime, brv, berr = times[0], brv[0], errors[0]

    # statistic can return nan in some cases
    if remove_nans:
        n = np.isnan(btime)
        btime = btime[~n]
        brv = brv[~n]
        if err is not None:
            berr = berr[~n]

    if err is not None:
        if n_consecutive and consecutive_step:
            return btime, brv, berr, random_choices
        return btime, brv, berr

    else:
        if n_consecutive and consecutive_step:
            return btime, brv, random_choices
        return btime, brv


def bin_ccf_mask(time, ccf_mask):
    logger = setup_logger()
    indices = binRV(time, None, binning_indices=True)
    indices = np.r_[indices, time.size]
    bmask = []
    for i1, i2 in zip(indices, indices[1:]):
        um = np.unique(ccf_mask[i1:i2]).squeeze()
        if um.size > 1:
            logger.error(f'Non-unique CCF mask within one night (t={time[i1]:.1f}). '
                         'Setting to NaN, but RV should be discarded')
            bmask.append('nan')
        else:
            bmask.append(um)
    return np.array(bmask)


import os
from glob import glob
import pickle
import numpy as np
import matplotlib.pyplot as plt

from .setup_logger import logger

from tqdm import tqdm
import astropy.units as u, astropy.constants as const
from astropy.io import fits

def doppler_shift(wave: np.ndarray, flux: np.ndarray, velocity: float):
    """ Doppler shift a spectrum by a given velocity

    Args:
        wave (np.ndarray): wavelength array
        flux (np.ndarray): flux array
        velocity (float): velocity in km/s
    """
    c = const.c.to(u.km/u.second).value
    doppler_factor = np.sqrt((1 + velocity/c) / (1 - velocity/c))
    new_wavelength = wave * doppler_factor
    new_flux = np.interp(new_wavelength, wave, flux)
    return new_wavelength, new_flux

def fit_gaussian_to_line(wave, flux, center_wavelength, around=0.15 * u.angstrom,
                         careful_continuum=False, plot=True, ax=None):
    from scipy.optimize import curve_fit
    if center_wavelength < wave.min() or center_wavelength > wave.max():
        raise ValueError('`center_wavelength` is outside the wavelength range')
    if center_wavelength < wave[np.nonzero(flux)].min() or center_wavelength > wave[np.nonzero(flux)].max():
        raise ValueError('`center_wavelength` is outside the wavelength range where flux is not zero')

    try:
        wave <<= u.angstrom
    except u.UnitConversionError as e:
        raise ValueError(f'could not convert `wave` to Angstroms: {e}') from None

    try:
        center_wavelength <<= u.angstrom
    except u.UnitConversionError as e:
        raise ValueError(f'could not convert `center_wavelength` to Angstroms: {e}') from None

    try:
        around <<= u.angstrom
    except u.UnitConversionError as e:
        raise ValueError(f'could not convert `around` to Angstroms: {e}') from None


    def gaussian(x, amp, cen, wid, off):
        return amp * np.exp(-(x-cen)**2 / (2*wid**2)) + off

    wave_around = (wave > center_wavelength - around) & (wave < center_wavelength + around)
    w, f = wave[wave_around].value, flux[wave_around]

    if careful_continuum:
        wave_around_continuum = (wave > center_wavelength - 10*around) & (wave < center_wavelength + 10*around)
        wc, fc = wave[wave_around_continuum].value, flux[wave_around_continuum]
        lim = np.percentile(fc, 80)
        wc = wc[fc > lim]
        fc = fc[fc > lim]
        w, f = np.r_[wc, w], np.r_[fc, f]
        f = f[np.argsort(w)]
        w = np.sort(w)

    lower, upper = np.array([
        [-np.inf, 0], 
        [-np.inf, np.inf], 
        [0.0, 0.11], 
        [0.9*f.max(), 1.1*f.max()]
    ]).T

    try:
        popt, pcov = curve_fit(gaussian, w, f, p0=[-np.ptp(f), center_wavelength.value, 0.1, f.max()],
                               bounds=(lower, upper))
    except RuntimeError as e:
        logger.warning(f'fit_gaussian_to_line: {e}')
        return None, np.nan, np.nan

    EW = A = (np.sqrt(2) * np.abs(popt[0]) * np.abs(popt[2]) * np.sqrt(np.pi)) / popt[3]
    perr = np.sqrt(np.diag(pcov))
    EW_err = (np.sqrt(2) * np.abs(perr[0]) * np.abs(perr[2]) * np.sqrt(np.pi)) / perr[3]

    if plot:
        if ax is None:
            fig, ax = plt.subplots(figsize=(8, 4))
        if careful_continuum:
            ax.plot(w, f, 'ko', ms=4, zorder=1)
            wave_around_plot = wave_around_continuum
        else:
            ax.plot(wave[wave_around], flux[wave_around], 'ko', ms=4, zorder=1)
            ax.plot(wave[wave_around], flux[wave_around] - gaussian(w, *popt), 'o', ms=2)
            wave_around_plot = (wave > center_wavelength - 2*around) & (wave < center_wavelength + 2*around)
        # ax.plot(wave[wave_around_plot], flux[wave_around_plot], 'o', ms=2)
        w = wave[wave_around_plot].value
        ax.plot(w, gaussian(w, *popt), 'r-')
        ax.fill_between([popt[1]-A, popt[1]+A], popt[3]+popt[0], popt[3],
                         color='C2', alpha=0.1, lw=0)

    return popt, EW*1e3, EW_err*1e3

def detrend(w, f):
    if w.shape[0] > w.shape[1]:
        w = np.copy(w).T
        f = np.copy(f).T

    f_detrended = np.zeros_like(f)
    for i, (ww, ff) in enumerate(zip(w, f)):
        m = np.nonzero(ff)
        fit = np.polyval(np.polyfit(ww[m] - np.median(ww[m]), ff[m], 1), ww - np.median(ww[m]))
        f_detrended[i] = ff - fit + np.median(ff[m])
    return w, f_detrended

def build_master(self, limit=None, plot=True):
    files = sorted(glob(f'{self.star}_downloads/*S1D_A.fits'))
    if self.verbose:
        logger.info(f'found {len(files)} S1D files')

    files = files[:limit]

    if len(files) == 0:
        if self.verbose:
            logger.warning('Should probably run `download_s1d` first')
        return

    if plot:
        fig, axs = plt.subplots(2, 1, figsize=(9, 6), sharex=True, constrained_layout=True)
        for ax in axs:
            ax.set(xlabel=r'wavelength air [$\AA$]', ylabel='flux')
        axs[0].set_title(self.star, loc='right', fontsize=10)

    w0 = fits.getdata(files[0])['wavelength_air']
    master_flux = np.zeros_like(w0)
    for file in files:
        rv = fits.getval(file, '*CCF RV')[0]
        flux = fits.getdata(file)['flux']
        _, new_flux = doppler_shift(w0, flux, rv)
        master_flux += new_flux
        if plot:
            axs[0].plot(w0, new_flux, alpha=0.5)

    master_flux /= len(files)
    if plot:
        axs[1].plot(w0, master_flux, 'k', label='master')
        axs[1].legend()
        axs[0].legend([], [], title=f'{len(files)} S1D spectra')

    return w0, master_flux


def determine_stellar_parameters(self, linelist: str, plot=True, **kwargs):
    try:
        from juliacall import Main as jl
        jl.seval("using Korg")
        Korg = jl.Korg
    except ModuleNotFoundError:
        msg = 'this function requires juliacall and Korg.jl, please `pip install juliacall`'
        logger.error(msg)
        return

    w, f = build_master(self, plot=plot)

    linelist = np.genfromtxt(linelist, dtype=None, encoding=None, names=True)
    lines = [
        Korg.Line(line['wl'], line['loggf'], Korg.Species(line['elem'].replace('Fe', 'Fe ')), line['EP'])
        for line in linelist
    ]

    if self.verbose:
        logger.info(f'found {len(lines)} lines in linelist')
        logger.info('measuring EWs...')

    EW = []
    pbar = tqdm(linelist)
    for line in pbar:
        pbar.set_description(f'{line["elem"]} {line["wl"]}')
        _, ew, _ = fit_gaussian_to_line(w, f, line['wl'], plot=plot,
                                        careful_continuum=kwargs.pop('careful_continuum', False))
        EW.append(ew)

    lines = list(np.array(lines)[~np.isnan(EW)])
    EW = np.array(EW)[~np.isnan(EW)]

    if self.verbose:
        logger.info('determining stellar parameters (can take a few minutes)...')

    callback = lambda p, r, A: print('current parameters:', p)
    result = Korg.Fit.ews_to_stellar_parameters(lines, EW, callback=callback)
    par, stat_err, sys_err = result

    if self.verbose:
        logger.info(f'best-fit stellar parameters:')
        logger.info(f'  Teff: {par[0]:.0f} ± {sys_err[0]:.0f} K')
        logger.info(f'  logg: {par[1]:.2f} ± {sys_err[1]:.2f} dex')
        logger.info(f'  m/H :  {par[3]:.2f} ± {sys_err[3]:.2f} dex')

    r = {
        'teff': (par[0], sys_err[0]),
        'logg': (par[1], sys_err[1]),
        'vmic': (par[2], sys_err[2]),
        'moh': (par[3], sys_err[3]),
    }

    with open(f'{self.star}_stellar_parameters.pkl', 'wb') as f:
        pickle.dump(r, f)
    
    return r
